# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Webglviewer(Component):
    """A Webglviewer component.
ExampleComponent is an example component.
It takes a property, `label`, and
displays it.
It renders an input with the property `value`
which is editable by the user.

Keyword arguments:
- id (string; required): The ID used to identify this component in Dash callbacks.
- data (list; optional): A grid or a list of grids that will be rendered when this component is rendered.
- idx (number; optional): Index of datum in data list which should be rendered.
- sky (list; optional): Sky map, expects an array of six arrays.
Value have to be between 0 and 255.
Each array represents one side of the cube map (flattened pixels of an image).
Expects 4 channels (r,g,b,a).
Image has to be quadratic.
- material_type (string; optional): Material type used for rendering.
Possible types: SMOKE, DARK_SMOKE, LIGHT_SMOKE, SOLID, LIQUID
- representation_type (string; optional): Specifies the representation type used.
Possible types: DENSITY, SDF, PARTICLE
- scale (number; optional): Particle scale."""
    @_explicitize_args
    def __init__(self, id=Component.REQUIRED, data=Component.UNDEFINED, idx=Component.UNDEFINED, sky=Component.UNDEFINED, material_type=Component.UNDEFINED, representation_type=Component.UNDEFINED, scale=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'data', 'idx', 'sky', 'material_type', 'representation_type', 'scale']
        self._type = 'Webglviewer'
        self._namespace = 'webglviewer'
        self._valid_wildcard_attributes = []
        self.available_properties = ['id', 'data', 'idx', 'sky', 'material_type', 'representation_type', 'scale']
        self.available_wildcard_properties = []

        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}

        for k in ['id']:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(Webglviewer, self).__init__(**args)



import warnings
import six

from .physics import Physics, State, struct, _ChainedPhysics, _as_physics


class StateCollection(dict):

    def __init__(self, states=None):
        """
Create a state collection from a dictionary of states.
        :param states: dict mapping from state names to states
        :type states: dict or list or tuple None
        """
        if states is None:
            states = {}
        elif not isinstance(states, dict):
            states = {state.name: state for state in states}
        dict.__init__(self, states)

    def __setitem__(self, key, val):
        raise AttributeError('StateCollections are immutable')

    def all_with_tag(self, tag):
        return [s for s in self.values() if tag in s.tags]

    def all_instances(self, cls):
        return [s for s in self.values() if isinstance(s, cls)]

    def state_added(self, state):
        assert state.name not in self, 'A state with name "%s" is already present. Use state_replaced() to replace it.' % state.name
        new_states = self.copy()
        new_states[state.name] = state
        return StateCollection(new_states)

    def state_replaced(self, new_state):
        assert new_state.name in self, 'No state found with name "%s"' % new_state.name
        new_states = dict(self)
        new_states[new_state.name] = new_state
        return StateCollection(new_states)

    def state_removed(self, state):
        name = state if isinstance(state, six.string_types) else state.name
        new_states = dict(self)
        del new_states[name]
        return StateCollection(new_states)

    def find(self, name):
        warnings.warn("StateCollection.find is deprecated. Use statecollection[name] instead.", DeprecationWarning)
        return dict.__getitem__(self, name)

    def __getitem__(self, item):
        if isinstance(item, State):
            return self[item.name]
        if isinstance(item, six.string_types):
            return dict.__getitem__(self, item)
        if struct.isstruct(item):
            return struct.map(lambda x: self[x], item, content_type=struct.INVALID)
        try:
            return self[item.name]
        except AttributeError as e:
            pass
        raise ValueError('Illegal argument: %s' % item)

    def __getattr__(self, item):
        return self[item]

    def default_physics(self):
        warnings.warn("StateCollection will be removed in the future.", DeprecationWarning)
        return CollectivePhysics()

    def __repr__(self):
        return '[' + ', '.join((str(s) for s in self)) + ']'

    def __contains__(self, item):
        if isinstance(item, State):
            return item.name in self
        if isinstance(item, six.string_types):
            return dict.__contains__(self, item)
        raise ValueError('Illegal type: %s' % type(item))

    def __hash__(self):
        return 0

    @property
    def states(self):
        return self

    def copied_with(self, **kwargs):
        if len(kwargs) == 0:
            return self
        assert len(kwargs) == 1
        name, value = next(iter(kwargs.items()))
        assert name == 'states'
        return StateCollection(value)

    @property
    def shape(self):
        return StateCollection({name: state.shape for name, state in self.items()})

    @property
    def staticshape(self):
        return StateCollection({name: state.staticshape for name, state in self.items()})

    @property
    def dtype(self):
        return StateCollection({name: state.dtype for name, state in self.items()})


CollectiveState = StateCollection


class CollectivePhysics(Physics):

    def __init__(self):
        Physics.__init__(self, {})
        self.physics = {}  # map from name to Physics

    def step(self, state_collection, dt=1.0, **dependent_states):
        assert len(dependent_states) == 0
        if len(state_collection) == 0:
            return state_collection
        unhandled_states = list(state_collection.values())
        next_states = {}
        partial_next_state_collection = StateCollection(next_states)

        for sweep in range(len(state_collection)):
            for state in tuple(unhandled_states):
                physics = self.for_(state)
                if self._all_dependencies_fulfilled(physics.blocking_dependencies, state_collection, partial_next_state_collection):
                    next_state = self.substep(state, state_collection, dt, partial_next_state_collection=partial_next_state_collection)
                    assert next_state is not None, "step() called on %s returned None for state '%s'" % (type(physics).__name__, state)
                    assert isinstance(next_state, State), "step() called on %s dit not return a State but '%s' for state '%s'" % (type(physics).__name__, next_state, state)
                    assert next_state.name == state.name, "The state name must remain constant during step(). Caused by '%s' on state '%s'." % (type(physics).__name__, state)
                    next_states[next_state.name] = next_state
                    unhandled_states.remove(state)
            partial_next_state_collection = StateCollection(next_states)
            if len(unhandled_states) == 0:
                ordered_states = [partial_next_state_collection[state] for state in state_collection]
                return StateCollection(ordered_states)

        # Error
        errstr = 'Cyclic blocking_dependencies in simulation: %s' % unhandled_states
        for state in tuple(unhandled_states):
            physics = self.for_(state)
            state_dict = self._gather_dependencies(physics.blocking_dependencies, state_collection, {})
            errstr += '\nState "%s" with physics "%s" depends on %s' % (state, physics, state_dict)
        raise AssertionError(errstr)

    def substep(self, state, state_collection, dt, override_physics=None, partial_next_state_collection=None):
        physics = self.for_(state) if override_physics is None else override_physics
        # --- gather dependencies
        dependent_states = {}
        self._gather_dependencies(physics.dependencies, state_collection, dependent_states)
        if partial_next_state_collection is not None:
            self._gather_dependencies(physics.blocking_dependencies, partial_next_state_collection, dependent_states)
        # --- execute step ---
        next_state = physics.step(state, dt, **dependent_states)
        return next_state

    def _gather_dependencies(self, dependencies, state_collection, result_dict):
        for statedependency in dependencies:
            if statedependency.state_name is not None:
                matching_states = state_collection.find(statedependency.state_name)
            else:
                matching_states = state_collection.all_with_tag(statedependency.tag)
            if statedependency.single_state:
                assert len(matching_states) == 1, 'Dependency %s requires 1 state but found %d' % (statedependency, len(matching_states))
                value = matching_states[0]
            else:
                value = tuple(matching_states)
            result_dict[statedependency.parameter_name] = value
        return result_dict

    def _all_dependencies_fulfilled(self, dependencies, all_states, computed_states):
        state_dict = self._gather_dependencies(dependencies, all_states, {})
        for name, states in state_dict.items():
            if isinstance(states, tuple):
                for state in states:
                    if state.name not in computed_states:
                        return False
            else:  # single state
                if states.name not in computed_states:
                    return False
        return True

    def for_(self, state):
        return self.physics[state.name] if state.name in self.physics else state.default_physics()

    def add(self, name, physics):
        self.physics[name] = _as_physics(physics)

    def remove(self, name):
        if name in self.physics:
            del self.physics[name]



import numbers

import numpy
from phi import math, struct

from .field import Field, propagate_flags_children
from .flag import DIVERGENCE_FREE


@struct.definition()
class ConstantField(Field):

    def __init__(self, value=1.0, name=None, batch_size=1, **kwargs):
        data = _convert_constant_to_data(value)
        Field.__init__(self, **struct.kwargs(locals(), ignore='value'))

    def sample_at(self, points):
        return _expand_axes(self.data, points, self._batch_size)

    @property
    def rank(self):
        return None

    @property
    def component_count(self):
        return self.data.shape[-1]

    def unstack(self):
        flags = propagate_flags_children(self.flags, self.rank, 1)
        return [ConstantField(c, '%s[%d]' % (self.name, i), flags=flags, batch_size=self._batch_size) for i, c in enumerate(math.unstack(self.data, -1))]

    @property
    def points(self):
        return None

    def compatible(self, other_field):
        return True

    def __repr__(self):
        return repr(self.data)


def _convert_constant_to_data(value):
    if isinstance(value, numbers.Number):
        value = math.to_float(math.expand_dims(value))
    if isinstance(value, (list, tuple)):
        value = math.to_float(numpy.array(value))
    if len(math.staticshape(value)) < 2:
        value = math.expand_dims(value)
    return value


def _expand_axes(data, points, batch_size=1):
    assert math.spatial_rank(data) >= 0
    data = math.expand_dims(data, 1, math.spatial_rank(points) - math.spatial_rank(data))
    points_axes = math.staticshape(points)[1:-1]
    data_axes = math.staticshape(data)[1:-1]
    for d_points, d_data in zip(points_axes, data_axes):
        assert d_points % d_data == 0
    tilings = [batch_size or 1] + [d_points // d_data for d_points, d_data in zip(math.staticshape(points)[1:-1], math.staticshape(data)[1:-1])] + [1]
    data = math.tile(data, tilings)
    return data



import numpy as np

from phi import math, struct

from . import Physics
from .domain import DomainState
from .field import AnalyticField


@struct.definition()
class Pattern(DomainState):

    def __init__(self, domain, u=0, v=0, tags=('pattern',), **kwargs):
        DomainState.__init__(self, **struct.kwargs(locals()))

    @struct.variable(default=0, dependencies=DomainState.domain)
    def u(self, U):
        return self.centered_grid('U', U, dtype=np.float32)

    @struct.variable(default=0, dependencies=DomainState.domain)
    def v(self, V):
        return self.centered_grid('V', V, dtype=np.float32)

    @struct.constant(default=0.19)
    def du(self, du):
        return du

    @struct.constant(default=0.05)
    def dv(self, dv):
        return dv

    @struct.constant(default=0.04)
    def f(self, f):
        return f

    @struct.constant(default=0.05)
    def k(self, k):
        return k


class ReactionDiffusion(Physics):

    def __init__(self):
        Physics.__init__(self)

    def step(self, pattern, dt=1.0, **kwargs):
        # if struct.all(Material.periodic(pattern.domain.boundaries)):
        #     dx = math.mean(pattern.domain.dx)
        #     lu = math.fourier_laplace(pattern.u) / dx ** 2
        #     lv = math.fourier_laplace(pattern.v) / dx ** 2
        # else:
        lu = pattern.u.laplace()
        lv = pattern.v.laplace()
        uvv = pattern.u * pattern.v ** 2
        su = pattern.du * lu - uvv + pattern.f * (1 - pattern.u)
        sv = pattern.dv * lv + uvv - (pattern.f + pattern.k) * pattern.v
        return pattern.copied_with(u=pattern.u + dt * su, v=pattern.v + dt * sv)


@struct.definition()
class Seed(AnalyticField):

    def __init__(self, center=(0, 0), size=0, mode='RANDOM', factor=1.0, name='Seed', data=1.0, **kwargs):
        rank = math.staticshape(center)[-1]
        AnalyticField.__init__(self, **struct.kwargs(locals()))

    @struct.constant()
    def center(self, center):
        return center

    @struct.constant()
    def size(self, size):
        return size

    @struct.constant()
    def factor(self, factor):
        return factor

    @struct.constant()
    def mode(self, mode):
        return mode

    def sample_at(self, points):
        if self.mode == 'EXP':
            envelope = math.exp(-0.5 * math.sum((points - self.center) ** 2, axis=-1, keepdims=True) / self.size ** 2)
            envelope = math.to_float(envelope)
            return envelope * self.factor

        elif self.mode == 'RECT':
            conf = np.zeros(points.shape)
            conf[:, self.center[0] - self.size:self.center[0] + self.size, self.center[1] - self.size:self.center[1] + self.size, :] = np.ones(conf[:, self.center[0] - self.size:self.center[0] + self.size, self.center[1] - self.size:self.center[1] + self.size, :].shape)
            return conf[:, :, :, :-1] * self.factor

        elif self.mode == 'RANDOM':

            conf = np.random.random_sample(points.shape)
            conf[:, 0, :, :] *= 0
            conf[:, -1, :, :] *= 0
            conf[:, :, 0, :] *= 0
            conf[:, :, -1, :] *= 0

            return conf[:, :, :, :-1] * self.factor



import numpy as np
from phi import math, struct

from . import Physics, StateDependency
from .domain import DomainState
from .field import AnalyticField, GeometryMask, union_mask
from .field.effect import ADD, FieldEffect, effect_applied


@struct.definition()
class QuantumWave(DomainState):

    def __init__(self, domain, amplitude=1, mass=0.1, tags=('qwave',), **kwargs):
        DomainState.__init__(self, **struct.kwargs(locals()))

    @struct.variable(default=1, dependencies=DomainState.domain)
    def amplitude(self, amplitude):
        return self.centered_grid('amplitude', amplitude, dtype=np.complex64)

    @struct.constant(default=0.1)
    def mass(self, mass):
        return mass

    def default_physics(self):
        return SCHROEDINGER


def normalize_probability(probability_amplitude):
    p = math.to_float(abs(probability_amplitude) ** 2)
    P = math.sum(p, math.spatial_dimensions(p), keepdims=True)
    return probability_amplitude / math.to_complex(math.sqrt(P))


psquare = math.abs_square


class Schroedinger(Physics):

    def __init__(self, margin=1):
        Physics.__init__(self, [StateDependency('obstacles', 'obstacle'),
                                StateDependency('potentials', 'potential_effect', blocking=True)])
        self.margin = margin

    def step(self, state, dt=1.0, potentials=(), obstacles=()):
        if len(potentials) == 0:
            potential = 0
        else:
            potential = math.zeros_like(math.real(state.amplitude))  # for the moment, allow only real potentials
            for pot in potentials:
                potential = effect_applied(pot, potential, dt)
            potential = potential.data

        amplitude = state.amplitude.data

        # Rotate by potential
        rotation = math.exp(1j * math.to_complex(potential * dt))
        amplitude = amplitude * rotation

        # Move by rotating in Fourier space
        amplitude_fft = math.fft(amplitude)
        laplace = math.fftfreq(state.resolution, mode='square')
        amplitude_fft *= math.exp(-1j * (2 * np.pi)**2 * math.to_complex(dt) * laplace / (2 * state.mass))
        amplitude = math.ifft(amplitude_fft)

        obstacle_mask = union_mask([obstacle.geometry for obstacle in obstacles]).at(state.amplitude).data
        amplitude *= 1 - obstacle_mask

        normalized = False
        symmetric = False
        if not symmetric:
            boundary_mask = math.zeros(state.domain.centered_shape(1, batch_size=1)).data
            boundary_mask[[slice(None)] + [slice(self.margin,-self.margin) for i in math.spatial_dimensions(boundary_mask)] + [slice(None)]] = 1
            amplitude *= boundary_mask

        if len(obstacles) > 0 or not symmetric:
            amplitude = normalize_probability(amplitude)
            normalized = True

        return state.copied_with(amplitude=amplitude)


SCHROEDINGER = Schroedinger()


def StepPotential(geometry, height):
    return FieldEffect(GeometryMask(geometry, name='potential') * height, ['potential'], mode=ADD)


@struct.definition()
class WavePacket(AnalyticField):

    def __init__(self, center, size, wave_vector, name='wave_packet', data=1.0, **kwargs):
        rank = math.staticshape(center)[-1]
        AnalyticField.__init__(self, **struct.kwargs(locals()))

    @struct.constant()
    def center(self, center): return center

    @struct.constant()
    def size(self, size): return size

    @struct.constant()
    def wave_vector(self, wave_vector):
        if len(math.shape(wave_vector)) == 0:
            wave_vector = math.expand_dims(wave_vector, 0)
        return wave_vector

    def sample_at(self, points):
        envelope = math.exp(-0.5 * math.sum((points - self.center) ** 2, axis=-1, keepdims=True) / self.size ** 2)
        envelope = math.to_float(envelope)
        wave = math.exp(1j * math.to_float(math.expand_dims(np.dot(points, self.wave_vector), -1))) * envelope
        return wave * self.data


@struct.definition()
class HarmonicPotential(AnalyticField):

    def __init__(self, center, unit_distance, maximum_value=1.0, data=1.0, name='harmonic', **kwargs):
        rank = math.shape(center)[-1]
        AnalyticField.__init__(self, **struct.kwargs(locals()))

    @struct.constant()
    def center(self, center): return center

    @struct.constant()
    def unit_distance(self, distance): return distance

    @struct.constant()
    def maximum_value(self, maximum_value): return maximum_value

    def sample_at(self, points):
        x = (points - self.center) / self.unit_distance
        pot = math.sum(x ** 2, -1, keepdims=True) * self.data
        if self.maximum_value is not None:
            pot = math.minimum(pot, self.maximum_value)
        return math.to_float(pot)


@struct.definition()
class SinPotential(AnalyticField):

    def __init__(self, k, phase_offset=0, data=1.0, name='harmonic', **kwargs):
        rank = math.size(k)
        AnalyticField.__init__(self, **struct.kwargs(locals()))

    @struct.variable()
    def k(self, k):
        """ Wave vector. Determines wave length and direction. """
        return k

    @struct.variable()
    def phase_offset(self, phase_offset): return phase_offset

    def sample_at(self, x):
        phase_offset = math.batch_align_scalar(self.phase_offset, 0, x)
        k = math.batch_align(self.k, 1, x)
        data = math.batch_align(self.data, 1, x)
        spatial_phase = math.sum(k * x, -1, keepdims=True)
        result = math.sin(math.to_float(spatial_phase + phase_offset)) * math.to_float(data)
        return result

    def __repr__(self):
        return 'Sin(x*%s)' % self.k

    @property
    def component_count(self):
        return math.staticshape(self.data)[-1]



# Because division is different in Python 2 and 3
from __future__ import division

import numpy as np

from phi import struct
from phi.backend.dynamic_backend import DYNAMIC_BACKEND as math
from phi.struct.functions import mappable

from .helper import (_contains_axis, _dim_shifted, _get_pad_width,
                     _get_pad_width_axes, all_dimensions, rank,
                     spatial_dimensions, spatial_rank)


def indices_tensor(tensor, dtype=None):
    """
    Returns an index tensor of the same spatial shape as the given tensor.
    Each index denotes the location within the tensor starting from zero.
    Indices are encoded as vectors in the index tensor.

    :param tensor: a tensor of shape (batch size, spatial dimensions..., component size)
    :param dtype: NumPy data type or `None` for default
    :return: an index tensor of shape (1, spatial dimensions..., spatial rank)
    """
    spatial_dimensions = list(tensor.shape[1:-1])
    idx_zyx = np.meshgrid(*[range(dim) for dim in spatial_dimensions], indexing='ij')
    idx = np.stack(idx_zyx, axis=-1).reshape([1, ] + spatial_dimensions + [len(spatial_dimensions)])
    if dtype is not None:
        return idx.astype(dtype)
    else:
        return math.to_float(idx)


def normalize_to(target, source=1, epsilon=1e-5, batch_dims=1):
    """
    Multiplies the target so that its total content matches the source.

    :param target: a tensor
    :param source: a tensor or number
    :param epsilon: small number to prevent division by zero or None.
    :return: normalized tensor of the same shape as target
    """
    target_total = math.sum(target, axis=tuple(range(batch_dims, math.ndims(target))), keepdims=True)
    denominator = math.maximum(target_total, epsilon) if epsilon is not None else target_total
    source_total = math.sum(source, axis=tuple(range(batch_dims, math.ndims(source))), keepdims=True)
    return target * (source_total / denominator)


def batch_align(tensor, innate_dims, target, convert_to_same_backend=True):
    if isinstance(tensor, (tuple, list)):
        return [batch_align(t, innate_dims, target) for t in tensor]
    # --- Convert type ---
    if convert_to_same_backend:
        backend = math.choose_backend([tensor, target])
        tensor = backend.as_tensor(tensor)
        target = backend.as_tensor(target)
    # --- Batch align ---
    ndims = len(math.staticshape(tensor))
    if ndims <= innate_dims:
        return tensor  # There is no batch dimension
    target_ndims = len(math.staticshape(target))
    assert target_ndims >= ndims
    if target_ndims == ndims:
        return tensor
    return math.expand_dims(tensor, axis=(-innate_dims - 1), number=(target_ndims - ndims))


def batch_align_scalar(tensor, innate_spatial_dims, target):
    if rank(tensor) == 0:
        assert innate_spatial_dims == 0
        return math.expand_dims(tensor, 0, len(math.staticshape(target)))
    if math.staticshape(tensor)[-1] != 1 or math.ndims(tensor) <= 1:
        tensor = math.expand_dims(tensor, -1)
    result = batch_align(tensor, innate_spatial_dims + 1, target)
    return result


def blur(field, radius, cutoff=None, kernel="1/1+x"):
    """
Warning: This function can cause NaN in the gradients, reason unknown.

Runs a blur kernel over the given tensor.
    :param field: tensor
    :param radius: weight function curve scale
    :param cutoff: kernel size
    :param kernel: Type of blur kernel (str). Must be in ('1/1+x', 'gauss')
    :return:
    """
    if cutoff is None:
        cutoff = min(int(round(radius * 3)), *field.shape[1:-1])

    xyz = np.meshgrid(*[range(-int(cutoff), (cutoff) + 1) for _ in field.shape[1:-1]])
    d = math.to_float(np.sqrt(np.sum([x**2 for x in xyz], axis=0)))
    if kernel == "1/1+x":
        weights = math.to_float(1) / (d / radius + 1)
    elif kernel.lower() == "gauss":
        weights = math.exp(- d / radius / 2)
    else:
        raise ValueError("Unknown kernel: %s" % kernel)
    weights /= math.sum(weights)
    weights = math.reshape(weights, list(weights.shape) + [1, 1])
    return math.conv(field, weights)


def l1_loss(tensor, batch_norm=True, reduce_batches=True):
    if struct.isstruct(tensor):
        all_tensors = struct.flatten(tensor)
        return sum(l1_loss(tensor, batch_norm, reduce_batches) for tensor in all_tensors)
    if reduce_batches:
        total_loss = math.sum(math.abs(tensor))
    else:
        total_loss = math.sum(math.abs(tensor), axis=list(range(1, len(tensor.shape))))
    if batch_norm and reduce_batches:
        batch_size = math.shape(tensor)[0]
        return math.div(total_loss, math.to_float(batch_size))
    else:
        return total_loss


def l2_loss(tensor, batch_norm=True):
    return l_n_loss(tensor, 2, batch_norm=batch_norm)


def l_n_loss(tensor, n, batch_norm=True):
    if struct.isstruct(tensor):
        all_tensors = struct.flatten(tensor)
        return sum(l_n_loss(tensor, n, batch_norm) for tensor in all_tensors)
    total_loss = math.sum(tensor ** n) / n
    if batch_norm:
        batch_size = math.shape(tensor)[0]
        return math.div(total_loss, math.to_float(batch_size))
    else:
        return total_loss


def frequency_loss(tensor, frequency_falloff=100, reduce_batches=True):
    """
    Instead of minimizing each entry of the tensor, minimize the frequencies of the tensor, emphasizing lower frequencies over higher ones.

    :param reduce_batches: whether to reduce the batch dimension of the loss by adding the losses along the first dimension
    :param tensor: typically actual - target
    :param frequency_falloff: large values put more emphasis on lower frequencies, 1.0 weights all frequencies equally.
    :return: scalar loss value
    """
    if struct.isstruct(tensor):
        all_tensors = struct.flatten(tensor)
        return sum(frequency_loss(tensor, frequency_falloff, reduce_batches) for tensor in all_tensors)
    diff_fft = abs_square(math.fft(tensor))
    k = fftfreq(tensor.shape[1:-1], mode='absolute')
    weights = math.exp(-0.5 * k ** 2 * frequency_falloff ** 2)
    return l1_loss(diff_fft * weights, reduce_batches=reduce_batches)


@mappable()
def abs_square(complex):
    return math.imag(complex) ** 2 + math.real(complex) ** 2


# Divergence

def divergence(tensor, dx=1, difference='central'):
    """
    Computes the spatial divergence of a vector channel from finite differences.

    :param tensor: vector field; tensor of shape (batch size, spatial dimensions..., spatial rank)
    :param dx: distance between adjacent grid points (default 1)
    :param difference: type of difference, one of ('forward', 'central') (default 'forward')
    :return: tensor of shape (batch size, spatial dimensions..., 1)
    """
    assert difference in ('central', 'forward', 'backward'), difference
    rank = spatial_rank(tensor)
    if difference == 'forward':
        return _divergence_nd(tensor, (0, 1)) / dx ** rank
    elif difference == 'backward':
        return _divergence_nd(tensor, (-1, 0)) / dx ** rank
    else:
        return _divergence_nd(tensor, (-1, 1)) / (2 * dx) ** rank


def _divergence_nd(tensor, relative_shifts):
    rank = spatial_rank(tensor)
    tensor = math.pad(tensor, _get_pad_width(rank, (-relative_shifts[0], relative_shifts[1])))
    components = []
    for dimension in range(rank):
        lower, upper = _dim_shifted(tensor, dimension, relative_shifts, diminish_others=(-relative_shifts[0], relative_shifts[1]), components=rank - dimension - 1)
        components.append(upper - lower)
    return math.sum(components, 0)


# Gradient

def gradient(tensor, dx=1, difference='forward', padding='replicate'):
    """
    Calculates the gradient of a scalar channel from finite differences.
    The gradient vectors are in reverse order, lowest dimension first.

    :param tensor: channel with shape (batch_size, spatial_dimensions..., 1)
    :param dx: physical distance between grid points (default 1)
    :param difference: type of difference, one of ('forward', 'backward', 'central') (default 'forward')
    :param padding: tensor padding mode
    :return: tensor of shape (batch_size, spatial_dimensions..., spatial rank)
    """
    assert tensor.shape[-1] == 1, "Gradient requires a scalar channel as input"
    assert 1 not in tensor.shape[1:-1], "All spatial dimensions must have size larger than 1, got %s" % tensor.shape
    if difference.lower() == 'central':
        return _gradient_nd(tensor, padding, (-1, 1)) / (dx * 2)
    elif difference.lower() == 'forward':
        return _gradient_nd(tensor, padding, (0, 1)) / dx
    elif difference.lower() == 'backward':
        return _gradient_nd(tensor, padding, (-1, 0)) / dx
    else:
        raise ValueError('Invalid difference type: {}. Can be CENTRAL or FORWARD'.format(difference))


def _gradient_nd(tensor, padding, relative_shifts):
    rank = spatial_rank(tensor)
    tensor = math.pad(tensor, _get_pad_width(rank, (-relative_shifts[0], relative_shifts[1])), mode=padding)
    components = []
    for dimension in range(rank):
        lower, upper = _dim_shifted(tensor, dimension, relative_shifts, diminish_others=(-relative_shifts[0], relative_shifts[1]))
        components.append(upper - lower)
    return math.concat(components, axis=-1)


def axis_gradient(tensor, spatial_axis):
    dims = range(spatial_rank(tensor))
    upper_slices = tuple([(slice(1, None) if i == spatial_axis else slice(None)) for i in dims])
    lower_slices = tuple([(slice(-1) if i == spatial_axis else slice(None)) for i in dims])
    diff = tensor[(slice(None),) + upper_slices + (slice(None),)] \
        - tensor[(slice(None),) + lower_slices + (slice(None),)]
    return diff


# Laplace

@mappable()
def laplace(tensor, padding='replicate', axes=None, use_fft_for_periodic=False):
    """
    Spatial Laplace operator as defined for scalar fields.
    If a vector field is passed, the laplace is computed component-wise.

    :param use_fft_for_periodic: If True and padding='circular', uses FFT to compute laplace
    :param tensor: n-dimensional field of shape (batch, spacial dimensions..., components)
    :param padding: 'valid', 'constant', 'reflect', 'replicate', 'circular'
    :param axes: The second derivative along these axes is summed over
    :type axes: list
    :return: tensor of same shape
    """
    rank = spatial_rank(tensor)
    if padding is None or padding == 'valid':
        pass  # do not pad tensor
    elif padding in ('circular', 'wrap') and use_fft_for_periodic:
        return fourier_laplace(tensor)
    else:
        tensor = math.pad(tensor, _get_pad_width_axes(rank, axes, val_true=[1, 1], val_false=[0, 0]), padding)
    # --- convolutional laplace ---
    if axes is not None:
        return _sliced_laplace_nd(tensor, axes)
    if rank == 2:
        return _conv_laplace_2d(tensor)
    elif rank == 3:
        return _conv_laplace_3d(tensor)
    else:
        return _sliced_laplace_nd(tensor)


def _conv_laplace_2d(tensor):
    kernel = math.to_float([[0., 1., 0.], [1., -4., 1.], [0., 1., 0.]])
    kernel = kernel.reshape((3, 3, 1, 1))
    if tensor.shape[-1] == 1:
        return math.conv(tensor, kernel, padding='VALID')
    else:
        return math.concat([math.conv(tensor[..., i:i + 1], kernel, padding='VALID') for i in range(tensor.shape[-1])], -1)


def _conv_laplace_3d(tensor):
    """
    3D/Cube laplace stencil in 3D+2D [3,3,3,1,1]
    array([[[[[ 0.]], [[ 0.]], [[ 0.]]],
            [[[ 0.]], [[ 1.]], [[ 0.]]],
            [[[ 0.]], [[ 0.]], [[ 0.]]]],
           [[[[ 0.]], [[ 1.]], [[ 0.]]],
            [[[ 1.]], [[-6.]], [[ 1.]]],
            [[[ 0.]], [[ 1.]], [[ 0.]]]],
           [[[[ 0.]], [[ 0.]], [[ 0.]]],
            [[[ 0.]], [[ 1.]], [[ 0.]]],
            [[[ 0.]], [[ 0.]], [[ 0.]]]]]
    returns ...

    padding explicitly done in laplace(), hence here not needed
    """
    kernel = math.to_float([[[0., 0., 0.], [0., 1., 0.], [0., 0., 0.]],
                            [[0., 1., 0.], [1., -6., 1.], [0., 1., 0.]],
                            [[0., 0., 0.], [0., 1., 0.], [0., 0., 0.]]])
    kernel = kernel.reshape((3, 3, 3, 1, 1))
    if tensor.shape[-1] == 1:
        return math.conv(tensor, kernel, padding='VALID')
    else:
        return math.concat([math.conv(tensor[..., i:i + 1], kernel, padding='VALID')
                            for i in range(tensor.shape[-1])], -1)


def _sliced_laplace_nd(tensor, axes=None):
    """
    Laplace Stencil for N-Dimensions
    aggregated from (c)enter, (u)pper, and (l)ower parts
    """
    rank = spatial_rank(tensor)
    dims = range(rank)
    components = []
    for ax in dims:
        if _contains_axis(axes, ax, rank):
            lower, center, upper = _dim_shifted(tensor, ax, (-1, 0, 1), diminish_others=(1, 1), diminish_other_condition=lambda other_ax: _contains_axis(axes, other_ax, rank))
            components.append(upper + lower - 2 * center)
    return math.sum(components, 0)


@mappable()
def fourier_laplace(tensor, times=1):
    """
Applies the spatial laplce operator to the given tensor with periodic boundary conditions.

*Note:* The results of `fourier_laplace` and `laplace` are close but not identical.

This implementation computes the laplace operator in Fourier space.
The result for periodic fields is exact, i.e. no numerical instabilities can occur, even for higher-order derivatives.
    :param tensor: tensor, assumed to have periodic boundary conditions
    :param times: number of times the laplace operator is applied. The computational cost is independent of this parameter.
    :return: tensor of same shape as `tensor`
    """
    frequencies = math.fft(math.to_complex(tensor))
    k = fftfreq(math.staticshape(tensor)[1:-1], mode='square')
    fft_laplace = -(2 * np.pi)**2 * k
    return math.real(math.ifft(frequencies * fft_laplace ** times))


@mappable()
def fourier_poisson(tensor, times=1):
    """ Inverse operation to `fourier_laplace`. """
    frequencies = math.fft(math.to_complex(tensor))
    k = fftfreq(math.staticshape(tensor)[1:-1], mode='square')
    fft_laplace = -(2 * np.pi)**2 * k
    fft_laplace[(0,) * math.ndims(k)] = np.inf
    inv_fft_laplace = 1 / fft_laplace
    inv_fft_laplace[(0,) * math.ndims(k)] = 0
    return math.cast(math.real(math.ifft(frequencies * inv_fft_laplace**times)), math.dtype(tensor))

def fftfreq(resolution, mode='vector', dtype=None):
    """
    Returns the discrete Fourier transform sample frequencies.
    These are the frequencies corresponding to the components of the result of `math.fft` on a tensor of shape `resolution`.

    :param resolution: grid resolution measured in cells
    :param mode: one of (None, 'vector', 'absolute', 'square')
    :param dtype: data type of the returned tensor
    :return: tensor holding the frequencies of the corresponding values computed by math.fft
    """
    assert mode in ('vector', 'absolute', 'square')
    k = np.meshgrid(*[np.fft.fftfreq(int(n)) for n in resolution], indexing='ij')
    k = math.expand_dims(math.stack(k, -1), 0)
    if dtype is not None:
        k = k.astype(dtype)
    else:
        k = math.to_float(k)
    if mode == 'vector':
        return k
    k = math.sum(k**2, axis=-1, keepdims=True)
    if mode == 'square':
        return k
    else:
        return math.sqrt(k)


# Downsample / Upsample

def downsample2x(tensor, interpolation='linear', axes=None):
    if struct.isstruct(tensor):
        return struct.map(lambda s: downsample2x(s, interpolation, axes),
                          tensor, recursive=False)

    if interpolation.lower() != 'linear':
        raise ValueError('Only linear interpolation supported')
    rank = spatial_rank(tensor)
    if axes is None:
        axes = range(rank)
    tensor = math.pad(tensor,
                      [[0, 0]]
                      + [([0, 1] if (dim % 2) != 0 and _contains_axis(axes, ax, rank) else [0, 0]) for ax, dim in enumerate(tensor.shape[1:-1])]
                      + [[0, 0]], 'replicate')
    for axis in axes:
        upper_slices = tuple([(slice(1, None, 2) if i == axis else slice(None)) for i in range(rank)])
        lower_slices = tuple([(slice(0, None, 2) if i == axis else slice(None)) for i in range(rank)])
        tensor_sum = tensor[(slice(None),) + upper_slices + (slice(None),)] + tensor[(slice(None),) + lower_slices + (slice(None),)]
        tensor = tensor_sum / 2
    return tensor


def upsample2x(tensor, interpolation='linear'):
    if struct.isstruct(tensor):
        return struct.map(lambda s: upsample2x(s, interpolation), tensor, recursive=False)

    if interpolation.lower() != 'linear':
        raise ValueError('Only linear interpolation supported')
    dims = range(spatial_rank(tensor))
    vlen = tensor.shape[-1]
    spatial_dims = tensor.shape[1:-1]
    rank = spatial_rank(tensor)
    tensor = math.pad(tensor, _get_pad_width(rank), 'replicate')
    for dim in dims:
        lower, center, upper = _dim_shifted(tensor, dim, (-1, 0, 1))
        combined = math.stack([0.25 * lower + 0.75 * center, 0.75 * center + 0.25 * upper], axis=2 + dim)
        tensor = math.reshape(combined, [-1] + [spatial_dims[dim] * 2 if i == dim else tensor.shape[i + 1] for i in dims] + [vlen])
    return tensor


def spatial_sum(tensor):
    summed = math.sum(tensor, axis=math.dimrange(tensor))
    for i in math.dimrange(tensor):
        summed = math.expand_dims(summed, i)
    return summed


def interpolate_linear(tensor, upper_weight, dimensions):
    """

    :param tensor:
    :param upper_weight: tensor of floats (leading dimensions must be 1) or nan to ignore interpolation along this axis
    :param dimensions: list or tuple of dimensions (first spatial axis=1) to be interpolated. Other axes are ignored.
    :return:
    """
    lower_weight = 1 - upper_weight
    for dimension in spatial_dimensions(tensor):
        if dimension in dimensions:
            upper_slices = tuple([(slice(1, None) if i == dimension else slice(None)) for i in all_dimensions(tensor)])
            lower_slices = tuple([(slice(-1) if i == dimension else slice(None)) for i in all_dimensions(tensor)])
            tensor = math.mul(tensor[upper_slices], upper_weight[..., dimension - 1]) + math.mul(tensor[lower_slices], lower_weight[..., dimension - 1])
    return tensor



"""
Surface material definitions including constants.
"""

import math

from phi import struct


@struct.definition()
class Material(struct.Struct):
    """
    Defines a surface material including the boundary conditions.
    """

    def __init__(self, name, **kwargs):
        struct.Struct.__init__(self, **struct.kwargs(locals()))

    @struct.constant()
    def name(self, name):
        """
Material name.
        """
        return str(name)

    @struct.constant(default=True)
    def solid(self, solid):
        """
Fluid can only enter non-solid cells or pass through non-solid boundaries.
        """
        assert isinstance(solid, bool)
        return solid

    @struct.derived()
    def open(self):
        return not self.solid and not self.periodic

    @struct.constant(default=0.0)
    def friction(self, friction):
        """
(only for solid materials) velocity decay rate in units of 1/time.
0: fluid can move parallell to the surface (no-stick),
1: fluid cannot move parallel (no-slip)
        """
        return friction

    def friction_multiplier(self, dt=1):
        """
Computes the velocity multiplication factor for fluid that moves along the surface for time dt.
        :param dt: time spent near surface (float)
        :return: factor (float)
        """
        if dt == 1 or self.friction == 1 or self.friction == 0:
            return 1 - self.friction
        else:
            time_friction_exponent = math.log(1 / (1 - self.friction))
            return math.exp(- dt * time_friction_exponent)

    @struct.constant(default=False)
    def periodic(self, periodic):
        """
Whether the boundary is periodic, i.e. seamlessly merges with the opposite end of the domain.
        """
        assert isinstance(periodic, bool)
        return periodic

    def __repr__(self):
        return self.name

    @struct.derived()
    def extrapolation_mode(self):
        """
Extrapolation mode for regular non-vector fields.
For vector fields that respect boundaries (e.g. velocity), use vector_extrapolation_mode.
    :return: one of ('periodic', 'boundary', 'constant').
    :rtype: str
        """
        if self.periodic:
            return 'periodic'
        if self.solid:
            return 'boundary'
        else:
            return 'constant'

    @struct.derived()
    def accessible_extrapolation_mode(self):
        if self.periodic:
            return 'periodic'
        if self.solid:
            return 'constant'
        else:
            return 'boundary'

    @struct.derived()
    def vector_extrapolation_mode(self):
        if self.periodic:
            return 'periodic'
        if self.solid:
            assert self.friction in (0, 1)
            return 'boundary' if self.friction == 0 else 'constant'
        else:
            return 'constant'


OPEN = Material('open', solid=False)
CLOSED = NO_STICK = SLIPPERY = Material('slippery', solid=True, friction=0)
NO_SLIP = STICKY = Material('sticky', solid=True, friction=1)
PERIODIC = Material('periodic', solid=False, periodic=True)



from .stream import DerivedStream


class Augmentstream(DerivedStream):

    def __init__(self, aug_dimensions, field, affect_flags=(DATAFLAG_TRAIN,)):
        DerivedStream.__init__(self, [field])
        self.field = self.inputs[0]
        self.affect_flags = affect_flags
        self.aug_dimensions = range(aug_dimensions) if isinstance(aug_dimensions, int) else aug_dimensions
        self.augmentation_factor = 2 ** len(self.aug_dimensions)

    def affects(self, datasource):
        if not self.affect_flags:
            return True
        for flag in self.affect_flags:
            if flag in datasource.flags:
                return True
        return False

    def size(self, datasource, lookup=False):
        return self.field.size(datasource, lookup=lookup) * (self.augmentation_factor if self.affects(datasource) else 1)

    def shape(self, datasource):
        return self.field.shape(datasource)

    def get(self, datasource, indices):
        if not self.affects(datasource):
            return self.field.get(datasource, indices)
        else:
            return self._augment_interleave(datasource, indices)

    def _augment_interleave(self, datasource, indices):
        arrays = self.field.get(datasource, [i // self.augmentation_factor for i in indices])
        for array, index in zip(arrays, indices):
            perm = index % self.augmentation_factor
            if perm != 0:
                array = self.augment_single(array, perm)
            yield array

    def augment_single(self, array, perm):
        raise NotImplementedError()


class AxisFlip(Augmentstream):

    def __init__(self, flip_dimensions, field, flip_vectors=True, affect_flags=(DATAFLAG_TRAIN,)):
        Augmentstream.__init__(self, flip_dimensions, field, affect_flags)
        self.flip_vectors = flip_vectors

    def augment_single(self, array, perm):
        slices = [slice(None, None, -1) if d >= 1 and perm & 2 ** (d - 1) != 0 else slice(None) for d in range(len(array.shape))]
        array = array[slices]
        if self.flip_vectors and array.shape[-1] == len(array.shape) - 2:
            flipped_components = [len(array.shape) - d - 3 for d in self.aug_dimensions if perm & 2 ** (d) != 0]
            array[..., flipped_components] *= -1
        return array



import sys
import inspect
import os
import warnings

from phi.app.app import App


class AppDisplay(object):

    def __init__(self, app):
        """
Creates a display for the given app and initializes the configuration.
This method does not set up the display. It only sets up the AppDisplay object and returns as quickly as possible.
        :param app: app to be displayed, may not be prepared or be otherwise invalid at this point.
        """
        self.app = app
        self.config = {}

    def configure(self, config):
        # type: (dict) -> None
        """
Updates the GUI configuration.
This method may only be called while the GUI is not yet visible, i.e. before show() is called.
        :param config: Complete or partial GUI-specific configuration. dictionary mapping from strings to JSON serializable values
        """
        self.config.update(config)

    def get_configuration(self):
        # type: () -> dict
        """
Returns the current configuration of the GUI.
The returned dictionary may only contain serializable values and all keys must be strings.
The configuration can be passed to another instance of this class using set_configuration().
        :rtype: dict
        """
        return self.config

    def setup(self):
        """
Sets up all necessary GUI components.

The GUI can register callbacks with the app to be informed about app-state changes induced externally.
The app can be assumed to be prepared when this method is called.

This method is called after set_configuration() but before show()

The return value of this method will be returned by show(app).
        """
        pass

    def show(self, caller_is_main):
        # type: (bool) -> bool
        """
Displays the previously setup GUI.
This method is blocking and returns only when the GUI is hidden.

This method will always be called after setup().
        :param caller_is_main: True if the calling script is the __main__ module.
        :return: Whether the GUI was displayed
        :rtype: bool
        """
        return False

    def play(self):
        """
Called if AUTORUN is enabled.
If no Display is specified, App.run() is called instead.
        """
        self.app.play()


DEFAULT_DISPLAY_CLASS = None

if 'headless' not in sys.argv:
    try:
        from .dash.dash_gui import DashGui
        DEFAULT_DISPLAY_CLASS = DashGui
    except ImportError as import_error:
        warnings.warn('GUI is disabled because of missing dependencies: %s. To install all dependencies, run $ pip install phiflow[gui]' % import_error)


AUTORUN = 'autorun' in sys.argv


def show(app=None, **config):

    frame_records = inspect.stack()[1]
    calling_module = inspect.getmodulename(frame_records[1])
    python_file = os.path.basename(sys.argv[0])[:-3]

    if app is None:
        fitting_models = _find_subclasses_in_module(App, calling_module, [])
        assert len(fitting_models) == 1, 'show() called without model but detected %d possible classes: %s' % (len(fitting_models), fitting_models)
        app = fitting_models[0]

    if inspect.isclass(app) and issubclass(app, App):
        app = app()

    called_from_main = inspect.getmodule(app.__class__).__name__ == '__main__' or calling_module == python_file

    app.prepare()

    display = None
    if DEFAULT_DISPLAY_CLASS is not None:
        display = DEFAULT_DISPLAY_CLASS(app)
        display.configure(config)
        display.setup()
    # --- Autorun ---
    if AUTORUN:
        if display is None:
            app.info('Starting execution because autorun is enabled.')
            app.play()  # asynchronous call
        else:
            display.play()
    # --- Show ---
    if display is None:
        warnings.warn('show() has no effect because no display is available. To use the web interface, run $ pip install phiflow[gui]')
        return app
    else:
        return display.show(called_from_main)  # blocking call


def _find_subclasses_in_module(base_class, module_name, result_list):
    subclasses = base_class.__subclasses__()
    for subclass in subclasses:
        if subclass not in result_list:
            mod_name = os.path.basename(inspect.getfile(subclass))[:-3]
            if mod_name == module_name:
                result_list.append(subclass)
            _find_subclasses_in_module(subclass, module_name, result_list)
    return result_list



class Backend:

    def __init__(self, name, precision=32):
        self._name = name
        self.precision = precision

    @property
    def name(self):
        return self._name

    @property
    def precision(self):
        """
        If `precision` is an integer, any Backend method may convert floating point values to this precision, even if the input had a different precision.

        If `precision` is `None`, the output of math operations has the same precision as its inputs.
        """
        return self._precision

    @precision.setter
    def precision(self, precision):
        """
        If `precision` is an integer, any Backend method may convert floating point values to this precision, even if the input had a different precision.

        If `precision` is `None`, the output of math operations has the same precision as its inputs.

        :param precision: one of (16, 32, 64, None)
        """
        assert precision in (None, 16, 32, 64)
        self._precision = precision

    @property
    def has_fixed_precision(self):
        return self.precision is not None

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def matches_name(self, name):
        return self.name.lower() == name.lower()

    def is_applicable(self, values):
        for value in values:
            if self.is_tensor(value, only_native=True):
                return True
        return False

    # --- Abstract math functions ---

    def is_tensor(self, x, only_native=False):
        """
        An object is considered a native tensor by a backend if no internal conversion is required by backend methods.
        An object is considered a tensor (nativer or otherwise) by a backend if it is not a struct (e.g. tuple, list) and all methods of the backend accept it as a tensor argument.

        :param x: object to check
        :param only_native: If True, only accepts true native tensor representations, not Python numbers or others that are also supported as tensors
        :return: whether `x` is considered a tensor by this backend
        :rtype: bool
        """
        raise NotImplementedError()

    def as_tensor(self, x, convert_external=True):
        """
        Converts a tensor-like object to the native tensor representation of this backend.
        If x is a native tensor of this backend, it is returned without modification.
        If x is a Python number (numbers.Number instance), `convert_numbers` decides whether to convert it unless the backend cannot handle Python numbers.

        *Note:* There may be objects that are considered tensors by this backend but are not native and thus, will be converted by this method.

        :param x: tensor-like, e.g. list, tuple, Python number, tensor
        :param convert_external: if False and `x` is a Python number that is understood by this backend, this method returns the number as is. This can help prevent type clashes like int32 vs int64.
        :return: tensor representation of `x`
        """
        raise NotImplementedError()

    def copy(self, tensor, only_mutable=False):
        raise NotImplementedError()

    def equal(self, x, y):
        raise NotImplementedError()

    def random_uniform(self, shape, low=0, high=1):
        raise NotImplementedError(self)

    def random_normal(self, shape):
        raise NotImplementedError(self)

    def stack(self, values, axis=0):
        raise NotImplementedError(self)

    def concat(self, values, axis):
        raise NotImplementedError(self)

    def pad(self, value, pad_width, mode='constant', constant_values=0):
        """
        Pad a tensor with values as specified by `mode` and `constant_values`.
        :param value: tensor
        :param pad_width: 2D tensor specifying the number of values padded to the edges of each axis in the form [[before axis 0, after axis 0], ...] including batch and component axes.
        :param mode: can be specified for each face, options are 'constant', 'replicate', 'circular', 'symmetric', 'reflect'
        :param constant_values: used for out-of-bounds points if mode='constant'
        """
        raise NotImplementedError(self)

    def resample(self, inputs, sample_coords, interpolation='linear', boundary='constant', constant_values=0):
        """
        Interpolates a regular grid at the specified coordinates.
        :param inputs: grid data
        :param sample_coords: tensor of floating grid indices. The last dimension must match the dimensions of inputs. The first grid point of dimension i lies at position 0, the last at data.shape[i]-1.
        :param interpolation: only 'linear' is currently supported
        :param boundary: values to use for coordinates outside the grid, can be specified for each face, options are 'constant', 'replicate', 'circular', 'symmetric', 'reflect'
        :param constant_values: Value used for constant boundaries, can be specified for each face
        """
        raise NotImplementedError(self)

    def reshape(self, value, shape):
        raise NotImplementedError(self)

    def sum(self, value, axis=None, keepdims=False):
        raise NotImplementedError(self)

    def prod(self, value, axis=None):
        raise NotImplementedError(self)

    def divide_no_nan(self, x, y):
        """ Computes x/y but returns 0 if y=0. """
        raise NotImplementedError(self)

    def where(self, condition, x=None, y=None):
        raise NotImplementedError(self)

    def mean(self, value, axis=None, keepdims=False):
        raise NotImplementedError(self)

    def py_func(self, func, inputs, Tout, shape_out, stateful=True, name=None, grad=None):
        raise NotImplementedError(self)

    def range(self, start, limit=None, delta=1, dtype=None):
        raise NotImplementedError(self)

    def zeros_like(self, tensor):
        raise NotImplementedError(self)

    def ones_like(self, tensor):
        raise NotImplementedError(self)

    def dot(self, a, b, axes):
        raise NotImplementedError(self)

    def matmul(self, A, b):
        raise NotImplementedError(self)

    def einsum(self, equation, *tensors):
        raise NotImplementedError(self)

    def while_loop(self, cond, body, loop_vars, shape_invariants=None, parallel_iterations=10, back_prop=True,
                   swap_memory=False, name=None, maximum_iterations=None):
        raise NotImplementedError(self)

    def abs(self, x):
        raise NotImplementedError(self)

    def sign(self, x):
        raise NotImplementedError(self)

    def round(self, x):
        raise NotImplementedError(self)

    def ceil(self, x):
        raise NotImplementedError(self)

    def floor(self, x):
        raise NotImplementedError(self)

    def max(self, x, axis=None, keepdims=False):
        raise NotImplementedError(self)

    def min(self, x, axis=None, keepdims=False):
        raise NotImplementedError(self)

    def maximum(self, a, b):
        raise NotImplementedError(self)

    def minimum(self, a, b):
        raise NotImplementedError(self)

    def clip(self, x, minimum, maximum):
        raise NotImplementedError(self)

    def with_custom_gradient(self, function, inputs, gradient, input_index=0, output_index=None, name_base='custom_gradient_func'):
        raise NotImplementedError(self)

    def sqrt(self, x):
        raise NotImplementedError(self)

    def exp(self, x):
        raise NotImplementedError(self)

    def conv(self, tensor, kernel, padding='same'):
        raise NotImplementedError(self)

    def expand_dims(self, a, axis=0, number=1):
        raise NotImplementedError(self)

    def shape(self, tensor):
        raise NotImplementedError(self)

    def staticshape(self, tensor):
        raise NotImplementedError(self)

    def to_float(self, x, float64=False):
        """
Converts a tensor to floating point values.
If this Backend uses a fixed precision, the tensor will be converted to that precision.
Otherwise, non-float inputs are converted to float32 (unless `float64=True`).

If `x` is mutable and of the correct floating type, returns a copy of `x`.

To convert float tensors to the backend precision but leave non-float tensors untouched, use `Backend.as_tensor()`.
        :param x: tensor
        :param float64: deprecated. Set Backend.precision = 64 to use 64 bit operations.
        """
        raise NotImplementedError(self)

    def to_int(self, x, int64=False):
        raise NotImplementedError(self)

    def to_complex(self, x):
        raise NotImplementedError(self)

    def dimrange(self, tensor):
        return range(1, len(tensor.shape) - 1)

    def gather(self, values, indices):
        raise NotImplementedError(self)

    def gather_nd(self, values, indices, batch_dims=0):
        raise NotImplementedError(self)

    def flatten(self, x):
        return self.reshape(x, (-1,))

    def std(self, x, axis=None, keepdims=False):
        raise NotImplementedError(self)

    def boolean_mask(self, x, mask):
        raise NotImplementedError(self)

    def isfinite(self, x):
        raise NotImplementedError(self)

    def scatter(self, points, indices, values, shape, duplicates_handling='undefined'):
        """
        This method expects the first dimension of indices and values to be the batch dimension.
        The batch dimension need not be specified in the indices array.

        All indices must be non-negative and are expected to be within bounds. Otherwise the behaviour is undefined.

        :param indices:
        :param values:
        :param shape:
        :param duplicates_handling: one of ('undefined', 'add', 'mean', 'any', 'last', 'no duplicates')
        """
        raise NotImplementedError(self)

    def any(self, boolean_tensor, axis=None, keepdims=False):
        raise NotImplementedError(self)

    def all(self, boolean_tensor, axis=None, keepdims=False):
        raise NotImplementedError(self)

    def fft(self, x):
        """
        Computes the n-dimensional FFT along all but the first and last dimensions.

        :param x: tensor of dimension 3 or higher
        """
        raise NotImplementedError(self)

    def ifft(self, k):
        """
        Computes the n-dimensional inverse FFT along all but the first and last dimensions.

        :param k: tensor of dimension 3 or higher
        """
        raise NotImplementedError(self)

    def imag(self, complex):
        raise NotImplementedError(self)

    def real(self, complex):
        raise NotImplementedError(self)

    def cast(self, x, dtype):
        raise NotImplementedError(self)

    def sin(self, x):
        raise NotImplementedError(self)

    def cos(self, x):
        raise NotImplementedError(self)

    def dtype(self, array):
        raise NotImplementedError(self)

    def tile(self, value, multiples):
        """
Repeats the tensor along each axis the number of times given by multiples.
If `multiples` has more dimensions than `value`, these dimensions are added to `value` as outer dimensions.
        :param value: tensor
        :param multiples: tuple or list of integers
        :return: tile tensor
        """
        raise NotImplementedError(self)

    def sparse_tensor(self, indices, values, shape):
        raise NotImplementedError(self)

    # --- Math function with default implementation ---

    def ndims(self, tensor):
        return len(self.staticshape(tensor))

    def size(self, array):
        return self.prod(self.shape(array))

    def batch_gather(self, tensor, batches):
        if isinstance(batches, int):
            batches = [batches]
        return tensor[batches, ...]

    def unstack(self, tensor, axis=0, keepdims=False):
        if axis < 0:
            axis += len(tensor.shape)
        if axis >= len(tensor.shape) or axis < 0:
            raise ValueError("Illegal axis value")
        result = []
        for slice_idx in range(tensor.shape[axis]):
            if keepdims:
                component = tensor[tuple([slice(slice_idx, slice_idx + 1) if d == axis else slice(None) for d in range(len(tensor.shape))])]
            else:
                component = tensor[tuple([slice_idx if d == axis else slice(None) for d in range(len(tensor.shape))])]
            result.append(component)
        return tuple(result)

    def add(self, a, b):
        return self.as_tensor(a, convert_external=False) + self.as_tensor(b, convert_external=False)

    def sub(self, a, b):
        return self.as_tensor(a, convert_external=False) - self.as_tensor(b, convert_external=False)

    def mul(self, a, b):
        return self.as_tensor(a, convert_external=False) * self.as_tensor(b, convert_external=False)

    def div(self, numerator, denominator):
        return self.as_tensor(numerator, convert_external=False) / self.as_tensor(denominator, convert_external=False)

    def pow(self, base, exp):
        return self.as_tensor(base, convert_external=False) ** self.as_tensor(exp, convert_external=False)

    def mod(self, dividend, divisor):
        dividend_tensor = self.as_tensor(dividend, convert_external=False)
        return dividend_tensor % self.cast(self.as_tensor(divisor, convert_external=False), self.dtype(dividend_tensor))



from phi import struct, math
from ._geom import Geometry
from ._empty import NO_GEOMETRY
from ._transform import rotate
from ._box import bounding_box, AABox


@struct.definition()
class Union(Geometry):

    def __init__(self, geometries, **kwargs):
        Geometry.__init__(self, **struct.kwargs(locals()))

    @struct.constant()
    def geometries(self, geometries):
        assert len(geometries) > 0
        rank = geometries[0].rank
        for g in geometries[1:]:
            assert g.rank == rank or g.rank is None or rank is None
        return tuple(geometries)

    @property
    def rank(self):
        return self.geometries[0].rank

    def lies_inside(self, location):
        return math.any([geometry.lies_inside(location) for geometry in self.geometries], axis=0)

    def approximate_signed_distance(self, location):
        return math.min([geometry.approximate_signed_distance(location) for geometry in self.geometries], axis=0)

    @property
    def center(self):
        return self._bounding_box().center

    def bounding_radius(self):
        return self._bounding_box().bounding_radius()

    def bounding_half_extent(self):
        return self._bounding_box().bounding_half_extent()

    def _bounding_box(self):
        boxes = [bounding_box(g) for g in self.geometries]
        lower = math.min([b.lower for b in boxes], axis=0)
        upper = math.max([b.upper for b in boxes], axis=0)
        return AABox(lower, upper)

    def shifted(self, delta):
        return Union([geometry.shifted(delta) for geometry in self.geometries])

    def rotated(self, angle):
        return rotate(self, angle)


def union(*geometries):
    if len(geometries) == 1 and isinstance(geometries[0], (tuple, list)):
        geometries = geometries[0]
    if len(geometries) == 0:
        return NO_GEOMETRY
    else:
        base_geometries = ()
        for geometry in geometries:
            base_geometries += geometry.geometries if isinstance(geometry, Union) else (geometry,)
        return Union(base_geometries)


Geometry.__add__ = lambda g1, g2: union(g1, g2)




import dash_core_components as dcc
import dash_html_components as html

from phi.viz.dash.player_controls import STEP_BUTTON, STEP_COMPLETE
from .dash_app import DashApp
from dash.dependencies import Input, Output


def build_log(dashapp):
    assert isinstance(dashapp, DashApp)

    layout = html.Div([
        html.Button('Refresh', id='log-refresh'),
        html.Div(id='log-dump', style={'background-color': '#F0F0F0'}, children=[
        ]),
        dcc.Interval(id='initialize-log', interval=200, max_intervals=1)
    ])

    @dashapp.dash.callback(Output('log-dump', 'children'), [STEP_COMPLETE, Input('initialize-log', 'n_intervals'), Input('log-refresh', 'n_clicks')])
    def refresh_log(*args):
        try:
            log_file = dashapp.app.log_file
            with open(log_file, 'r') as stream:
                log_text = stream.read()
            paragraphs = log_text.split('\n')
            return [html.P(paragraph) for paragraph in paragraphs]
        except BaseException as exc:
            return 'Could not load log file: %s' % exc

    return layout



from numbers import Number

from phi import math
from phi.math.blas import conjugate_gradient
from phi.math.helper import _dim_shifted
from phi.physics.field import CenteredGrid
from .solver_api import PoissonDomain, PoissonSolver
from phi.physics.material import Material


class GeometricCG(PoissonSolver):

    def __init__(self, accuracy=1e-5, max_iterations=2000):
        """
Conjugate gradient solver that geometrically calculates laplace pressure in each iteration.
Unlike most other solvers, this algorithm is TPU compatible but usually performs worse than SparseCG.

Obstacles are allowed to vary between examples but the same number of iterations is performed for each example in one batch.

        :param accuracy: the maximally allowed error on the divergence channel for each cell
        :param max_iterations: integer specifying maximum conjugent gradient loop iterations or None for no limit
        :param autodiff:
        """
        PoissonSolver.__init__(self, 'Single-Phase Conjugate Gradient', supported_devices=('CPU', 'GPU', 'TPU'), supports_guess=True, supports_loop_counter=True, supports_continuous_masks=True)
        assert math.is_scalar(accuracy), 'invalid accuracy: %s' % accuracy
        self.accuracy = accuracy
        self.max_iterations = max_iterations

    def solve(self, divergence, domain, guess, enable_backprop):
        assert isinstance(domain, PoissonDomain)
        fluid_mask = domain.accessible_tensor(extend=1)
        extrapolation = Material.extrapolation_mode(domain.domain.boundaries)

        def apply_A(pressure):
            pressure = CenteredGrid(pressure, extrapolation=extrapolation)
            pressure_padded = pressure.padded([[1, 1]] * pressure.rank)
            return _weighted_sliced_laplace_nd(pressure_padded.data, weights=fluid_mask)

        return conjugate_gradient(divergence, apply_A, guess, self.accuracy, self.max_iterations, back_prop=enable_backprop)


def _weighted_sliced_laplace_nd(tensor, weights):
    if tensor.shape[-1] != 1:
        raise ValueError('Laplace operator requires a scalar channel as input')
    dims = range(math.spatial_rank(tensor))
    components = []
    for dimension in dims:
        lower_weights, center_weights, upper_weights = _dim_shifted(weights, dimension, (-1, 0, 1), diminish_others=(1, 1))
        lower_values, center_values, upper_values = _dim_shifted(tensor, dimension, (-1, 0, 1), diminish_others=(1, 1))
        diff = math.mul(upper_values, upper_weights * center_weights) + math.mul(lower_values, lower_weights * center_weights) + math.mul(center_values, - lower_weights - upper_weights)
        components.append(diff)
    return math.sum(components, 0)



from phi import math
from phi.physics.field import SampledField, ConstantField, StaggeredGrid, CenteredGrid
from .field import StaggeredSamplePoints, Field


def advect(field, velocity, dt):
    """
Advect `field` along the `velocity` vectors using the default advection method.
    :param field: any built-in Field
    :type field: Field
    :param velocity: any Field
    :type velocity: Field
    :param dt: time increment
    :return: Advected field of same type as `field`
    """
    if isinstance(field, SampledField):
        return runge_kutta_4(field, velocity, dt=dt)
    if isinstance(field, ConstantField):
        return field
    if isinstance(field, (CenteredGrid, StaggeredGrid)):
        return semi_lagrangian(field, velocity, dt=dt)
    raise NotImplementedError(field)


def semi_lagrangian(field, velocity_field, dt):
    """
    Semi-Lagrangian advection with simple backward lookup.

    :param field: Field to be advected
    :param velocity_field: vector field, need not be compatible with with `field`.
    :param dt: time increment
    :return: Field compatible with input field
    """
    try:
        x0 = field.points
        v = velocity_field.at(x0)
        x = x0 - v * dt
        data = field.sample_at(x.data)
        return field.with_data(data)
    except StaggeredSamplePoints:
        advected = [semi_lagrangian(component, velocity_field, dt) for component in field.unstack()]
        return field.with_data(advected)


def mac_cormack(field, velocity_field, dt, correction_strength=1.0):
    """
    MacCormack advection uses a forward and backward lookup to determine the first-order error of semi-Lagrangian advection.
    It then uses that error estimate to correct the field values.
    To avoid overshoots, the resulting value is bounded by the neighbouring grid cells of the backward lookup.

    :param correction_strength: the estimated error is multiplied by this factor before being applied. The case correction_strength=0 equals semi-lagrangian advection. Set lower than 1.0 to avoid oscillations.
    :param field: Field to be advected
    :param velocity_field: vector field, need not be compatible with `field`.
    :param dt: time increment
    :return: Field compatible with input field
    """
    try:
        x0 = field.points
        v = velocity_field.at(x0)
        x_bwd = x0 - v * dt
        x_fwd = x0 + v * dt
        field_semi_la = field.with_data(field.sample_at(x_bwd.data))  # semi-Lagrangian advection
        field_inv_semi_la = field.with_data(field_semi_la.sample_at(x_fwd.data))  # inverse semi-Lagrangian advection
        new_field = field_semi_la + correction_strength * 0.5 * (field - field_inv_semi_la)
        field_clamped = math.clip(new_field, *field.general_sample_at(x_bwd.data, 'minmax'))  # Address overshoots
        return field_clamped
    except StaggeredSamplePoints:
        advected = [mac_cormack(component, velocity_field, dt) for component in field.unstack()]
        return field.with_data(advected)


def runge_kutta_4(field, velocity, dt):
    """
Lagrangian advection of particles.
    :param field: SampledField with any number of components
    :type field: SampledField
    :param velocity: Vector field
    :type velocity: Field
    :param dt: time increment
    :return: SampledField with same data as `field` but advected points
    """
    assert isinstance(field, SampledField)
    assert isinstance(velocity, Field)
    points = field.points
    # --- Sample velocity at intermediate points ---
    vel_k1 = velocity.at(points)
    vel_k2 = velocity.at(points + 0.5 * dt * vel_k1)
    vel_k3 = velocity.at(points + 0.5 * dt * vel_k2)
    vel_k4 = velocity.at(points + dt * vel_k3)
    # --- Combine points with RK4 scheme ---
    new_points = points + dt * (1 / 6.) * (vel_k1 + 2 * (vel_k2 + vel_k3) + vel_k4)
    result = SampledField(new_points.data, field.data, mode=field.mode, point_count=field._point_count, name=field.name)
    return result



"""
Worlds are used to manage simulations consisting of multiple states.
A world uses a StateCollection with CollectivePhysics to resolve state dependencies.
Worlds also facilitate referencing states when performing a forward simulation.

A default World, called `world` is provided for convenience.
"""
import inspect
import warnings
from typing import TypeVar

import six

from .collective import StateCollection
from .field.effect import Gravity
from .physics import Physics, State, Static


class StateProxy(object):
    """
    StateProxy mirrors all data_dict of a state in an associated world.
    While State objects are generally immutable, StateProxy also implements setting any attribute of the state.
    When an attribute is set, a copy of the state with the new value replaces the old state in the world.
    This object then mirrors the values of the new state.

    After world.step() is invoked, all Proxies of that world will mirror the state after stepping.

    To reference the current immutable state of
    """

    def __init__(self, enclosing_world, state_name):
        self.world = enclosing_world
        self.state_name = state_name

    @property
    def state(self):
        """
Finds and returns the state in the referenced world that matches the StateProxy's state_name.
        :return: State
        """
        state = self.world.state[self.state_name]
        assert state is not None
        return state

    @state.setter
    def state(self, state):
        """
Replaces the State in the referenced world.
        """
        assert state.name == self.state_name
        self.world.state = self.world.state.state_replaced(state)

    @property
    def physics(self):
        """
Returns the Physics object used by the referenced world for the system.
If not specified manually, the default physics is returned.
        """
        physics = self.world.physics.for_(self.state)
        assert physics is not None
        return physics

    @physics.setter
    def physics(self, physics):
        """
Sets a specific Physics object for the system in the referenced world.
        :param physics: Physics
        """
        assert isinstance(physics, Physics)
        self.world.physics.add(self.state_name, physics)

    def step(self, dt=1.0, physics=None):
        """
Steps this system in the referenced world. Other states are not affected.
        :param dt: time increment
        :param physics: specific Physics object to use, defaults to self.physics
        """
        self.world.step(self, dt=dt, physics=physics)

    def __getattr__(self, item):
        """
Returns an attribute from the referenced State.
        :param item: attribute name
        """
        assert item not in ('world', 'state_name', 'physics', 'state')
        return getattr(self.state, item)

    def __setattr__(self, key, value):
        """
Changes the referenced state by replacing it in the referenced world.
        :param key: State attribute name
        :param value: new value
        """
        if key in ('world', 'state_name', 'physics', 'state'):
            object.__setattr__(self, key, value)
        else:
            self.state = self.state.copied_with(**{key:value})


# pylint: disable-msg = invalid-name
S = TypeVar('S', bound=State)


class World(object):
    """
    A world object defines a global state as well as a set of rules (Physics objects) that definition how the state evolves.

    The world manages dependencies among the contained simulations and provides convenience methods for creating proxies for specific simulations.

    The method world.step() evolves the whole state or optionally a specific state in time.
    """

    def __init__(self, batch_size=None, add_default_objects=True):
        # --- Insert object / create proxy shortcuts ---
        self._state = self.physics = self.observers = self.batch_size = None
        self.reset(batch_size, add_default_objects)

    def reset(self, batch_size=None, add_default_objects=True):
        """
Resets the world to the default configuration.
This removes all States and observers.
        :param batch_size: int or None
        :param add_default_objects: if True, adds defaults like Gravity
        """
        self._state = StateCollection()
        self.physics = self._state.default_physics()
        self.observers = set()
        self.batch_size = batch_size
        if add_default_objects:
            self.add(Gravity())

    @property
    def state(self):
        """
Returns the current state of the world.
        :return: StateCollection
        """
        return self._state

    @property
    def age(self):
        """
Alias for world.state.age
        """
        return self._state.age

    @state.setter
    def state(self, state):
        """
Sets the current state of the world and informs all observers.
        :param state: StateCollection
        """
        assert state is not None
        assert isinstance(state, StateCollection)
        self._state = state
        for observer in self.observers:
            observer(self)

    def step(self, state=None, dt=1.0, physics=None):
        """
        Evolves the current world state by a time increment dt.
        If state is provided, only that state is evolved, leaving the others untouched.
        The optional physics parameter can then be used to override the default physics.
        Otherwise, all states are evolved.

        Calling World.step resolves all dependencies among simulations and then calls Physics.step on each simulation to evolve the states.

        Invoking this method alters the world state. To to_field a copy of the state, use :func:`World.stepped <~world.World.stepped>` instead.
            :param state: State, StateProxy or None
            :param dt: time increment
            :param physics: Physics object for the state or None for default
            :return: evolved state if a specific state was provided
        """
        if state is None:
            if physics is None:
                physics = self.physics
            self.state = physics.step(self._state, dt)
            return self.state
        else:
            if isinstance(state, StateProxy):
                state = state.state
            s = self.physics.substep(state, self._state, dt, override_physics=physics)
            self.state = self._state.state_replaced(s)
            return s

    def stepped(self, state=None, dt=1.0, physics=None):
        """
Similar to step() but does not change the state of the world. Instead, the new state is returned.
        """
        if state is None:
            if physics is None:
                physics = self.physics
            return physics.step(self._state, None, dt)
        else:
            if isinstance(state, StateProxy):
                state = state.state
            return self.physics.substep(state, self._state, dt, override_physics=physics)

    def add(self, state, physics=None):
        # type: (S, Physics) -> S
        """
Adds a State to the world that will be stepped forward in time each time world.step() is invoked.
        :param state: State or list of States
        :param physics: (optional) Physics to use during world.step(). If a list was provided for `state`, a matching list must be given for `state`.
        :return: StateProxy referencing the current state of the added system. If world.state is updated (e.g. because world.step() was called), the StateProxy will refer to the updated values.
        """
        if isinstance(state, dict):
            raise ValueError('Cannot add dict to world. Maybe you meant world.add(**dict)?')
        if isinstance(state, (tuple, list)):
            assert isinstance(physics, (tuple, list))
            assert len(state) == len(physics)
            return [self.add(s, p) for s, p in zip(state, physics)]
        else:
            if physics is not None:
                self.physics.add(state.name, physics)
            elif state.default_physics() is not None and not isinstance(state.default_physics(), Static):
                warnings.warn('No physics provided to world.add(%s). In the future this will default to static physics' % state)
            self.state = self.state.state_added(state)
            return StateProxy(self, state.name)

    def add_all(self, *states):
        """
Add a collection of states to the system using world.add(state).
        """
        warnings.warn('World.add_all() is deprecated. Use World.add(list_of_states) instead.', DeprecationWarning)
        for state in states:
            self.add(state)

    def remove(self, obj):
        """
Remove a system or collection of systems from the world.
        :param obj: one of the following: State, state name, subclass of State, tuple or list thereof
        """
        if inspect.isclass(obj):
            states = self.state.all_instances(obj)
            self.remove(states)
        elif isinstance(obj, (tuple, list)):
            for state in obj:
                self.remove(state)
        else:
            key = obj if isinstance(obj, six.string_types) else obj.name
            self.state = self.state.state_removed(key)
            self.physics.remove(key)

    def get_physics(self, state):
        """
        Looks up the Physics object associated with a given State or StateProxy.
        If no Physics object was registered manually, the state.default_physics() object is used.

        :param state: State or StateProxy contained in this world
        :return: Physics
        """
        if isinstance(state, StateProxy):
            state = state.state
        return self.physics.for_(state)

    def __getattr__(self, item):
        if item in self.state:
            return StateProxy(self, item)
        else:
            return object.__getattribute__(self, item)


world = World()



import logging
import os

import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Output, Input
from dash.exceptions import PreventUpdate

from phi.viz.dash.dash_app import DashApp
from phi.viz.dash.player_controls import STEP_COUNT, parse_step_count


BENCHMARK_BUTTON = Input('benchmark-button', 'n_clicks')
PROFILE_BUTTON = Input('profile-button', 'n_clicks')

NO_BENCHMARK_TEXT = '*No benchmarks available.*'
NO_PROFILES_TEXT = '*No profiles available.*'


def build_benchmark(dashapp):
    assert isinstance(dashapp, DashApp)

    layout = html.Div([
        dcc.Markdown('## Benchmark'),
        html.Div([
            html.Button('Benchmark', id=BENCHMARK_BUTTON.component_id)
        ]),
        dcc.Markdown(children=NO_BENCHMARK_TEXT, id='run-statistics'),
    ])

    @dashapp.dash.callback(Output('run-statistics', 'children'), [BENCHMARK_BUTTON], [STEP_COUNT])
    def run_benchmark(n_clicks, step_count):
        step_count = parse_step_count(step_count, dashapp, default=1)
        if n_clicks is None:
            return NO_BENCHMARK_TEXT
        if dashapp.app.running:
            return '*Pause the app before starting a benchmark.*'
        # --- Run benchmark ---
        step_count, time_elapsed = dashapp.app.benchmark(step_count)
        output = '### Benchmark Results\n'
        if step_count != step_count:
            output += 'The benchmark was stopped prematurely.  \n'
        if dashapp.app.record_data or dashapp.app.record_images:
            output += 'Recording was disabled during benchmark.  \n'
        output += 'Finished %d steps in %.03f seconds.' % (step_count, time_elapsed)
        output += '  \n*Average*: %.04f seconds per step, %.02f steps per second.' % (
            time_elapsed / step_count, step_count / time_elapsed)
        return output

    return layout


def build_tf_profiler(dashapp):
    assert isinstance(dashapp, DashApp)

    layout = html.Div([
        dcc.Markdown('## TensorFlow Profiler'),
        html.Div([
            html.Button('Profile', id=PROFILE_BUTTON.component_id)
        ]),
        dcc.Markdown(children=NO_PROFILES_TEXT, id='profile-output'),
    ])

    @dashapp.dash.callback(Output('profile-output', 'children'), [PROFILE_BUTTON], [STEP_COUNT])
    def run_benchmark(n_clicks, step_count):
        step_count = parse_step_count(step_count, dashapp, default=1)
        if n_clicks is None:
            return NO_PROFILES_TEXT
        if dashapp.app.running:
            return '*Pause the app before starting a profiled run.*'
        # --- Profile ---
        with dashapp.app.session.profiler() as profiler:
            timeline_file = profiler.timeline_file
            step_count, time_elapsed = dashapp.app.benchmark(step_count)
        output = '### Profiling Results\n'
        if step_count != step_count:
            output += 'The profiling run was stopped prematurely.  \n'
        if dashapp.app.record_data or dashapp.app.record_images:
            output += 'Recording was disabled during benchmark.  \n'
        output += 'Finished %d steps in %.03f seconds.' % (step_count, time_elapsed)
        output += '  \n*Average*: %.04f seconds per step, %.02f steps per second.' % (time_elapsed / step_count, step_count / time_elapsed)
        output += '  \nProfile saved. Open  \n*chrome://tracing/*  \n and load file  \n *%s*' % timeline_file
        return output
    return layout


TENSORBOARD_STATUS = Input('tensorboard-status', 'children')


def build_tensorboard_launcher(dashapp):
    assert isinstance(dashapp, DashApp)

    layout = html.Div([
        html.Div(id='tensorboard-div'),
        dcc.Interval(id='tensorboard-init', interval=200, max_intervals=1),
        html.Div(style={'display': 'none'}, id=TENSORBOARD_STATUS.component_id),
    ])

    @dashapp.dash.callback(Output('tensorboard-div', 'children'), [Input('tensorboard-init', 'n_intervals'), TENSORBOARD_STATUS])
    def update(*_):
        if 'tensorboard_url' in dashapp.config:
            return html.A('TensorBoard', href=dashapp.config['tensorboard_url'], id='tensorboard-href')
        else:
            return html.Button('Launch TensorBoard', id='launch-tensorboard')

    @dashapp.dash.callback(Output(TENSORBOARD_STATUS.component_id, TENSORBOARD_STATUS.component_property), [Input('launch-tensorboard', 'n_clicks')])
    def launch_tensorboard(clicks):
        if clicks:
            logging.info('Launching TensorBoard...')
            logdir = dashapp.app.session.summary_directory
            import phi.tf.profiling as profiling
            url = profiling.launch_tensorboard(logdir, port=dashapp.config.get('tensorboard_port', None))
            dashapp.config['tensorboard_url'] = url
            logging.info('TensorBoard launched, URL: %s' % url)
            return 'running'
        else:
            raise PreventUpdate()

    return layout


def build_system_controls(dashapp):
    assert isinstance(dashapp, DashApp)

    layout = html.Div([
        dcc.Markdown('## Application'),
        html.Button('Terminate', id='exit-button')
    ])

    @dashapp.dash.callback(Output('exit-button', 'style'), [Input('exit-button', 'n_clicks')])
    def exit_application(n):
        if n:
            logging.info('DashGUI: Exiting...')
            os._exit(0)  # exit() does not work from Dash threads

    return layout



from .field import Field, StaggeredSamplePoints, IncompatibleFieldTypes
from .flag import Flag, DIVERGENCE_FREE, L2_NORMALIZED
from .constant import ConstantField
from .grid import CenteredGrid
from .staggered_grid import StaggeredGrid, unstack_staggered_tensor
from .sampled import SampledField
from .analytic import AnalyticField, SymbolicFieldBackend
from .mask import GeometryMask, mask, union_mask
from .noise import Noise
from . import advect
from . import manta
from .util import diffuse, data_bounds, staggered_curl_2d


from phi import math

math.DYNAMIC_BACKEND.add_backend(SymbolicFieldBackend(math.DYNAMIC_BACKEND), priority=True)



import numpy as np
import six

from phi import math, struct
from phi.backend.backend_helper import general_grid_sample_nd
from phi.geom import AABox, box
from phi.geom.geometry import assert_same_rank
from phi.math.helper import map_for_axes
from phi.physics.domain import Domain
from phi.physics.material import Material
from phi.struct.functions import mappable
from phi.struct.tensorop import collapse

from .field import Field, propagate_flags_children
from .flag import SAMPLE_POINTS


def _crop_for_interpolation(data, offset_float, window_resolution):
    offset = math.to_int(offset_float)
    slices = [slice(o, o + res + 1) for o, res in zip(offset, window_resolution)]
    data = data[tuple([slice(None)] + slices + [slice(None)])]
    return data


@struct.definition()
class CenteredGrid(Field):

    def __init__(self, data, box=None, extrapolation='boundary', name=None, **kwargs):
        """Create new CenteredGrid from array like data

        :param data: numerical values to be set as values of CenteredGrid (immutable)
        :type data: array-like
        :param box: numerical values describing the surrounding area of the CenteredGrid, defaults to None
        :type box: domain.box, optional
        :param extrapolation: set conditions for boundaries, defaults to 'boundary'
        :type extrapolation: str, optional
        :param name: give CenteredGrid a custom name (immutable), defaults to None
        :type name: string, optional
        """
        Field.__init__(self, **struct.kwargs(locals()))

    @staticmethod
    def sample(value, domain, batch_size=None, name=None):
        assert isinstance(domain, Domain)
        if isinstance(value, Field):
            assert_same_rank(value.rank, domain.rank, 'rank of value (%s) does not match domain (%s)' % (value.rank, domain.rank))
            if isinstance(value, CenteredGrid) and value.box == domain.box and np.all(value.resolution == domain.resolution):
                data = value.data
            else:
                point_field = CenteredGrid.getpoints(domain.box, domain.resolution)
                point_field._batch_size = batch_size
                data = value.at(point_field).data
        else:  # value is constant
            if callable(value):
                x = CenteredGrid.getpoints(domain.box, domain.resolution).copied_with(extrapolation=Material.extrapolation_mode(domain.boundaries), name=name)
                value = value(x)
                return value
            components = math.staticshape(value)[-1] if math.ndims(value) > 0 else 1
            data = math.add(math.zeros((batch_size,) + tuple(domain.resolution) + (components,)), value)
        return CenteredGrid(data, box=domain.box, extrapolation=Material.extrapolation_mode(domain.boundaries), name=name)

    @struct.variable()
    def data(self, data):
        if data is None:
            return None
        if isinstance(data, (tuple, list)):
            data = np.array(data)  # numbers or objects
        if self.content_type in (struct.shape, struct.staticshape):
            assert math.ndims(data) == 1
        else:
            if math.ndims(data) < 2:
                data = math.expand_dims(data, 0, number=2 - math.ndims(data))
        return data
    data.override(struct.staticshape, lambda self, data: (self._batch_size,) + math.staticshape(data)[1:])

    @property
    def resolution(self):
        if self.content_type in (struct.VALID, struct.INVALID):
            return math.as_tensor(math.staticshape(self.data)[1:-1])
        elif self.content_type in (struct.shape, struct.staticshape):
            return self.data[1:-1]
        else:
            raise AssertionError('Cannot compute resolution of invalid CenteredGrid (content type = %s)' % self.content_type)

    @struct.constant(dependencies=Field.data)
    def box(self, box):
        return AABox.to_box(box, resolution_hint=self.resolution)

    @struct.derived()
    def dx(self):
        return self.box.size / self.resolution

    @property
    def rank(self):
        return len(self.resolution)

    @struct.constant(default='boundary')
    def extrapolation(self, extrapolation):
        if extrapolation is None:
            return 'boundary'
        assert extrapolation in ('periodic', 'constant', 'boundary') or isinstance(extrapolation, (tuple, list)), extrapolation
        return collapse(extrapolation)

    @struct.constant(default=0.0)
    def extrapolation_value(self, value):
        return collapse(value)

    @struct.constant(default='linear')
    def interpolation(self, interpolation):
        assert interpolation == 'linear'
        return interpolation

    def sample_at(self, points):
        local_points = self.box.global_to_local(points)
        local_points = math.mul(local_points, math.to_float(self.resolution)) - 0.5
        resampled = math.resample(self.data, local_points, boundary=_pad_mode(self.extrapolation), interpolation=self.interpolation, constant_values=_pad_value(self.extrapolation_value))
        return resampled

    def general_sample_at(self, points, reduce):
        local_points = self.box.global_to_local(points)
        local_points = math.mul(local_points, math.to_float(self.resolution)) - 0.5
        result = general_grid_sample_nd(self.data, local_points, boundary=_pad_mode(self.extrapolation), constant_values=_pad_value(self.extrapolation_value), math=math.choose_backend([self.data, points]), reduce=reduce)
        return result

    def at(self, other_field):
        if self.compatible(other_field):
            return self
        if isinstance(other_field, CenteredGrid) and np.allclose(self.dx, other_field.dx):
            paddings = _required_paddings_transposed(self.box, self.dx, other_field.box)
            if math.sum(paddings) == 0:
                origin_in_local = self.box.global_to_local(other_field.box.lower) * self.resolution
                data = _crop_for_interpolation(self.data, origin_in_local, other_field.resolution)
                dimensions = self.resolution != other_field.resolution
                dimensions = [d for d in math.spatial_dimensions(data) if dimensions[d - 1]]
                data = math.interpolate_linear(data, origin_in_local % 1.0, dimensions)
                return CenteredGrid(data, other_field.box, name=self.name, batch_size=self._batch_size)
            elif math.sum(paddings) < 16:
                padded = self.padded(np.transpose(paddings).tolist())
                return padded.at(other_field)
        return Field.at(self, other_field)

    @property
    def component_count(self):
        if self.content_type in (struct.shape, struct.staticshape):
            return self.data[-1]
        else:
            return self.data.shape[-1]

    def unstack(self):
        flags = propagate_flags_children(self.flags, self.rank, 1)
        components = math.unstack(self.data, axis=-1, keepdims=True)
        return [CenteredGrid(component, box=self.box, flags=flags, batch_size=self._batch_size) for i, component in enumerate(components)]

    @property
    def points(self):
        if self.is_valid and SAMPLE_POINTS in self.flags:
            return self
        return CenteredGrid.getpoints(self.box, self.resolution)

    @property
    def elements(self):
        return box(center=self.points.data, size=self.dx)

    def compatible(self, other_field):
        if isinstance(other_field, (Domain, CenteredGrid)):
            if self.box != other_field.box:
                return False
            if self.rank != other_field.rank:
                return False
            for r1, r2 in zip(self.resolution, other_field.resolution):
                if r1 != r2 and r2 != 1 and r1 != 1:
                    return False
            return True
        if not other_field.has_points:
            return other_field.compatible(self)
        else:
            return False

    def __can_validate__(self):
        return self.content_type in (struct.INVALID, struct.shape, struct.staticshape)

    def __repr__(self):
        if self.is_valid:
            return 'Grid[%s(%d), size=%s, %s]' % ('x'.join([str(r) for r in self.resolution]), self.component_count, self.box.size, self.dtype.data)
        else:
            return struct.Struct.__repr__(self)

    def padded(self, widths):
        if isinstance(widths, int):
            widths = [[widths, widths]] * self.rank
        data = math.pad(self.data, [[0, 0]] + widths + [[0, 0]], _pad_mode(self.extrapolation), constant_values=_pad_value(self.extrapolation_value))
        w_lower, w_upper = np.transpose(widths)
        box = AABox(self.box.lower - w_lower * self.dx, self.box.upper + w_upper * self.dx)
        return self.copied_with(data=data, box=box)

    def axis_padded(self, axis, lower, upper):
        widths = [[lower, upper] if ax == axis else [0,0] for ax in range(self.rank)]
        return self.padded(widths)

    @staticmethod
    def getpoints(box, resolution):
        idx_zyx = np.meshgrid(*[np.linspace(0.5 / dim, 1 - 0.5 / dim, dim) for dim in resolution], indexing="ij")
        local_coords = math.to_float(math.expand_dims(math.stack(idx_zyx, axis=-1), 0))
        points = box.local_to_global(local_coords)
        return CenteredGrid(points, box, name='grid_centers(%s, %s)' % (box, resolution), flags=[SAMPLE_POINTS])

    def laplace(self, physical_units=True, axes=None):
        if not physical_units:
            data = math.laplace(self.data, padding=_pad_mode(self.extrapolation), axes=axes)
        else:
            if not self.has_cubic_cells:
                raise NotImplementedError('Only cubic cells supported.')
            laplace = math.laplace(self.data, padding=_pad_mode(self.extrapolation), axes=axes)
            data = laplace / self.dx[0] ** 2
        extrapolation = map_for_axes(_gradient_extrapolation, self.extrapolation, axes, self.rank)
        return self.copied_with(data=data, extrapolation=extrapolation, flags=())

    def gradient(self, physical_units=True, difference='central'):
        if not physical_units or self.has_cubic_cells:
            data = math.gradient(self.data, dx=np.mean(self.dx), difference=difference, padding=_pad_mode(self.extrapolation))
            return self.copied_with(data=data, extrapolation=_gradient_extrapolation(self.extrapolation), flags=())
        else:
            raise NotImplementedError('Only cubic cells supported.')

    @property
    def has_cubic_cells(self):
        return np.allclose(self.dx, np.mean(self.dx))

    def normalized(self, total, epsilon=1e-5):
        if isinstance(total, CenteredGrid):
            total = total.data
        normalize_data = math.normalize_to(self.data, total, epsilon)
        return self.with_data(normalize_data)

    @struct.derived()
    def frequencies(self):
        return self.with_data(math.fftfreq(self.resolution, mode='vector') / self.dx)

    @struct.derived()
    def squared_frequencies(self):
        return self.with_data(math.sum(self.frequencies.data ** 2, axis=-1, keepdims=True))

    def fft(self):
        return self.with_data(math.fft(self.data))

    @struct.derived()
    def abs(self):
        return self.with_data(math.abs(self.data))


def _required_paddings_transposed(box, dx, target, threshold=1e-5):
    lower = math.to_int(math.ceil(math.maximum(0, box.lower - target.lower) / dx - threshold))
    upper = math.to_int(math.ceil(math.maximum(0, target.upper - box.upper) / dx - threshold))
    return [lower, upper]


def _pad_mode(extrapolation):
    """ Inserts 'constant' padding for batch dimension and channel dimension. """
    if isinstance(extrapolation, six.string_types):
        return _pad_mode_str(extrapolation)
    else:
        return _pad_mode_str(['constant'] + list(extrapolation) + ['constant'])


def _pad_value(value):
    if math.is_tensor(value):
        return value
    else:
        return [0] + list(value) + [0]


@mappable()
def _pad_mode_str(extrapolation):
    """
Converts an extrapolation string (or struct of strings) to a string that can be passed to math functions like math.pad or math.resample.
    :param extrapolation: field extrapolation
    :return: padding mode, same type as extrapolation
    """
    return {'periodic': 'circular',
            'boundary': 'replicate',
            'constant': 'constant'}[extrapolation]


@mappable()
def _gradient_extrapolation(extrapolation):
    """
Given the extrapolation of a field, returns the extrapolation mode of the corresponding gradient field.
    :param extrapolation: string or struct of strings
    :return: same type as extrapolation
    """
    return {'periodic': 'periodic',
            'boundary': 'constant',
            'constant': 'constant'}[extrapolation]



from .staggered_grid import StaggeredGrid, unstack_staggered_tensor
from .grid import CenteredGrid
from phi import math
from phi import geom


def staggered_grid(tensor, name='manta_staggered'):
    tensor = tensor[...,::-1]  # manta: xyz, phiflow: zyx
    assert math.staticshape(tensor)[-1] == math.spatial_rank(tensor)
    return StaggeredGrid(tensor, name=name)


def centered_grid(tensor, name='manta_centered', crop_valid=False):
    if crop_valid:
        tensor = tensor[(slice(None),) + (slice(-1),) * math.spatial_rank(tensor) + (slice(None),)]
    return CenteredGrid(tensor, name=name)



from phi import math, struct
from phi.geom import GLOBAL_AXIS_ORDER

from .analytic import AnalyticField


@struct.definition()
class AngularVelocity(AnalyticField):

    def __init__(self, location, strength=1.0, **kwargs):
        AnalyticField.__init__(self, rank=None, **struct.kwargs(locals()))

    def sample_at(self, points):
        points_rank = math.spatial_rank(points)
        src_rank = math.spatial_rank(self.location)
        # --- Expand shapes to format (batch_size, points_dims..., src_dims..., channels) ---
        points = math.expand_dims(points, axis=-2, number=src_rank)
        src_location = math.expand_dims(self.location, axis=-3, number=points_rank)
        src_strength = math.expand_dims(self.strength, axis=-1)
        if math.ndims(src_strength) == 1:
            src_strength = math.expand_dims(src_strength, axis=-2)
        src_strength = math.expand_dims(src_strength, axis=-3, number=points_rank)
        src_axes = tuple(range(-2, -2 - src_rank, -1))
        # --- Compute distances and falloff ---
        distances = points - src_location
        if self.falloff is not None:
            falloff_value = self.falloff(distances)
            strength = src_strength * falloff_value
        else:
            strength = src_strength
        # --- Compute velocities ---
        if math.staticshape(points)[-1] == 2:  # Curl in 2D
            dist_1, dist_2 = math.unstack(distances, axis=-1)
            if GLOBAL_AXIS_ORDER.is_x_first:
                velocity = strength * math.stack([-dist_2, dist_1], axis=-1)
            else:
                velocity = strength * math.stack([dist_2, -dist_1], axis=-1)
        elif math.staticshape(points)[-1] == 3:  # Curl in 3D
            raise NotImplementedError('not yet implemented')
        else:
            raise AssertionError('Vector product not available in > 3 dimensions')
        velocity = math.sum(velocity, axis=src_axes)
        return velocity

    @property
    def component_count(self):
        return self.rank

    @struct.variable()
    def location(self, loc):
        loc = math.to_float(loc)
        assert math.staticshape(loc)[-1] in (2, 3)
        if math.ndims(loc) < 2:
            loc = math.expand_dims(loc, axis=0, number=2 - math.ndims(loc))
        return loc

    @struct.variable()
    def strength(self, strength):
        return math.to_float(strength)

    @struct.constant()
    def falloff(self, falloff):
        assert callable(falloff) or falloff is None
        return falloff

    @property
    def rank(self):
        return math.staticshape(self.location)[-1]



import warnings

from phi import struct
from .field import advect
from .field.effect import effect_applied
from .field.util import diffuse
from .domain import DomainState
from .physics import Physics, StateDependency


@struct.definition()
class BurgersVelocity(DomainState):

    def __init__(self, domain, velocity=0, viscosity=0.1, name='burgers', **kwargs):
        DomainState.__init__(self, **struct.kwargs(locals()))

    @struct.variable(dependencies=DomainState.domain, default=0)
    def velocity(self, velocity):
        return self.centered_grid('velocity', velocity, self.rank)

    @struct.constant(default=0.1)
    def viscosity(self, viscosity):
        return viscosity

    def __add__(self, other):
        return self.copied_with(velocity=self.velocity + other)


class Burgers(Physics):

    def __init__(self, default_viscosity=0.1, viscosity=None, diffusion_substeps=1, advection=advect.semi_lagrangian):
        Physics.__init__(self, [StateDependency('effects', 'velocity_effect', blocking=True)])
        if viscosity is not None:
            warnings.warn("Argument 'viscosity' is deprecated, use 'default_viscosity' instead.", DeprecationWarning)
            default_viscosity = viscosity
        self.default_viscosity = default_viscosity
        self.diffusion_substeps = diffusion_substeps
        self.advection = advection

    def step(self, v, dt=1.0, effects=()):
        if isinstance(v, BurgersVelocity):
            return v.copied_with(velocity=self.step_velocity(v.velocity, v.viscosity, dt, effects, self.diffusion_substeps), age=v.age + dt)
        else:
            return self.step_velocity(v, self.default_viscosity, dt, effects, self.diffusion_substeps)

    def step_velocity(self, v, viscosity, dt, effects, diffusion_substeps):
        v = diffuse(v, dt * viscosity, substeps=diffusion_substeps)
        v = self.advection(v, v, dt)
        for effect in effects:
            v = effect_applied(effect, v, dt)
        return v.copied_with(age=v.age + dt)



import warnings

from phi import struct, math, geom

from .field import Field
from .analytic import AnalyticField


@struct.definition()
class GeometryMask(AnalyticField):

    def __init__(self, geometries, antialias=False, value=None, **kwargs):
        if value is not None:
            warnings.warn('Passing a value to GeometryMask has no effect.')
        rank = geometries.rank if isinstance(geometries, geom.Geometry) else geometries[0].rank
        AnalyticField.__init__(self, rank, **struct.kwargs(locals(), ignore=['value', 'rank']))

    @struct.constant()
    def geometries(self, geometry):
        """ Alias for `geometry`. """
        if isinstance(geometry, (tuple, list)):
            geometry = geom.union(geometry)
        assert isinstance(geometry, geom.Geometry)
        return geometry

    geometry = geometries

    @struct.constant(default=False)
    def antialias(self, antialias):
        """
If False, field values are either 0 (outside) or 1 (inside) and the field is not differentiable w.r.t. the geometry.
If True, field values smoothly go from 0 to 1 at the surface and the field is differentiable w.r.t. the geometry.
        """
        assert antialias in (True, False)
        return antialias

    def approximate_mean_value_in(self, geometry):
        if not self.antialias:
            return Field.approximate_mean_value_in(self, geometry)
        else:
            return self.geometry.approximate_fraction_inside(geometry)

    def sample_at(self, points):
        return math.to_float(self.geometry.lies_inside(points))

    @property
    def component_count(self):
        return 1


mask = GeometryMask


def union_mask(geometries):
    warnings.warn("union_mask() is deprecated, use mask(union()) instead.", DeprecationWarning)
    return mask(geom.union(*geometries))



import math
from bisect import bisect_left

import six

try:
    # Python 3
    from collections.abc import Iterable
except ImportError:
    # Python 2.7
    from collections import Iterable
from sys import getsizeof

import numpy as np
from phi import struct

from .stream import DataStream, SourceStream

SKIP = 'skip'
WRAP = 'wrap'
CLIP = 'clip'


class BatchReader(object):

    def __init__(self, dataset, fields):
        self._dataset = dataset
        self._index = 0
        self._streams = []
        self._fields = fields
        self.streams = tuple(filter(lambda x: isinstance(x, DataStream) or isinstance(x, six.string_types), struct.flatten(fields)))
        self.stream_mask = struct.map(lambda x: x in self.streams, self._fields, content_type='stream_mask')
        for stream in self.streams:
            if isinstance(stream, DataStream):
                self._streams.append(stream)
            elif isinstance(stream, six.string_types):
                self._streams.append(SourceStream(stream))
            else:
                assert False
        self._cache = _BatchCache()
        self.indexcache = None
        self._dataset_changed()

    @property
    def dataset(self):
        return self._dataset

    def _get_batch(self, indices):
        data_list = self._cache.get(indices, self._load, add_to_cache=True)
        data = list_swap_axes(data_list)
        data_map = {self.streams[i]: data[i] for i in range(len(self._streams))}
        return struct.map(lambda x, is_stream: data_map[x] if is_stream else x, struct.zip([self._fields, self.stream_mask]), content_type=struct.INVALID)

    def _load(self, indices):
        result = []
        for index in indices:
            arrays = []
            for stream in self._streams:
                source, local_index = self.indexcache.get_source_and_local_index(index)
                data = stream.get(source, [local_index])
                array = next(iter(data))  # TODO group frames by source before calling get
                arrays.append(array)
            result.append(arrays)
        return result

    def __getitem__(self, item):
        if isinstance(item, int):
            return self._get_batch([item])
        elif isinstance(item, slice):
            stop = len(self) if item.stop is None else item.stop
            if stop < 0:
                stop = len(self) + stop
            start = 0 if item.start is None else item.start
            step = 1 if item.step is None else item.step
            return self._get_batch(list(range(start, stop, step)))
        elif isinstance(item, Iterable):
            return self._get_batch(item)
        else:
            raise ValueError("Illegal index: %s" % item)

    def __len__(self):
        return self._len

    def _dataset_changed(self):
        self._cache.clear()
        # Compute length
        if len(self._streams) == 0:
            self._len = 0
            self.indexcache = None
        else:
            stream = self._streams[0]
            self._len = np.sum([stream.size(source, lookup=True) for source in self._dataset.sources], dtype=np.int64)
            self.indexcache = _IndexCache(self._dataset.sources, self._streams[0])

    def all_batches(self, batch_size=1, last=CLIP, loop=False):
        return _AdaptiveBatchIterator(self, batch_size, last, loop)


class _IndexCache(object):

    def __init__(self, sources, stream, randomize_scene_order=False, randomize_frame_order=False, randomize_indices=True):
        if randomize_scene_order and not randomize_indices:
            # python2 doesnt yet support indexing via array, manually permutate
            self.sources = []
            p = np.random.permutation(len(sources))
            for i in range(len(sources)):
                self.sources.append(sources[p[i]])
            # replace sometime with original code: self.sources = sources[np.random.permutation(len(sources))]
        else:
            self.sources = sources
        self.datastream = stream
        self.accumulated_sizes = []

    def get_source_and_local_index(self, index):
        max_size = 0 if len(self.accumulated_sizes) == 0 else self.accumulated_sizes[-1]
        while index >= max_size:
            source = self.sources[len(self.accumulated_sizes)]
            source_size = self.datastream.size(source, lookup=True)
            max_size = max_size + source_size
            self.accumulated_sizes.append(max_size)
        pos = bisect_left(self.accumulated_sizes, index + 1)
        local_index = index if pos == 0 else index - self.accumulated_sizes[pos - 1]
        return self.sources[pos], local_index


class _BatchCache(object):

    def __init__(self, capacity=512 * 1024 * 1024):
        self._access_order = []  # most recent ones last
        self._data_by_stream_by_index = {}
        self._size = 0
        self.capacity = capacity

    def get(self, indices, lookup_function, add_to_cache=True):
        indices = np.array(indices)
        are_cached = np.array([index in self._data_by_stream_by_index for index in indices])
        uncached_indices = indices[~are_cached]
        cached_indices = indices[are_cached]

        if len(uncached_indices) > 0:
            uncached_data = lookup_function(uncached_indices)
            if add_to_cache:
                for index, data in zip(uncached_indices, uncached_data):
                    self.add(index, data)
        else:
            uncached_data = []

        for index in cached_indices:
            self._access_order.remove(index)
            self._access_order.append(index)

        result = []
        for index in indices:
            if index in cached_indices:
                result.append(self._data_by_stream_by_index[index])
            else:
                result.append(uncached_data[int(np.where(uncached_indices == index)[0])])
        return result

    def clear(self):
        self._access_order = []
        self._data_by_stream_by_index = {}
        self._size = 0

    def add(self, index, data):
        data_size = np.sum([getsizeof(array) for array in data])
        self.remove_old(self.capacity - data_size)
        self._access_order.append(index)
        assert index not in self._data_by_stream_by_index
        self._data_by_stream_by_index[index] = data
        self._size += data_size

    def remove_old(self, target_capacity):
        if target_capacity < 0:
            target_capacity = 0
        while self._size > target_capacity:
            index = self._access_order.pop(0)
            data = self._data_by_stream_by_index.pop(index)
            for array in data:
                self._size -= getsizeof(array)


class _AdaptiveBatchIterator(object):

    def __init__(self, batchreader, batch_size, last, loop):
        assert isinstance(batchreader, BatchReader)
        assert batch_size > 0
        assert last in (SKIP, CLIP, WRAP)
        assert isinstance(loop, bool)
        self.reader = batchreader
        self.batch_size = batch_size
        self.last = last
        self.loop = loop
        self.index = 0

    def __iter__(self):
        return self

    def __next__(self):
        start = self.index
        stop = start + self.batch_size

        if self.last == CLIP:
            stop = min(stop, len(self.reader))
            if start == stop:
                if not self.loop:
                    raise StopIteration()
                else:
                    start = 0
                    stop = min(self.batch_size, len(self.reader))
            indices = range(start, stop)
        elif self.last == SKIP:
            if stop > len(self.reader):
                start = 0
                stop = self.batch_size
                if stop > len(self.reader):
                    if not self.loop:
                        raise StopIteration()
                    else:
                        raise AssertionError("Looping iterator with 0 batches")
            indices = range(start, stop)
        elif self.last == WRAP:
            # Repeat first frames at the end
            indices = range(start, start + self.batch_size)
            indices = [i % len(self.reader) for i in indices]
            stop = (start + self.batch_size) % len(self.reader)
        else:
            raise AssertionError()

        batch = self.reader[indices]
        self.index = stop
        return batch

    next = __next__

    def __len__(self):
        assert not self.loop, "Looping iterator has no finite length"
        if self.last == SKIP:
            return len(self.reader) // self.batch_size
        if self.last == CLIP or self.last == WRAP:
            return int(math.ceil(float(len(self.reader)) / self.batch_size))


def list_swap_axes(list_in, concatenate=True):
    if len(list_in) == 0:
        return list_in
    result = []
    for i in range(len(list_in[0])):
        subresult = []
        for sublist in list_in:
            subresult.append(sublist[i])
        if concatenate:
            subresult = np.concatenate(subresult, axis=0)
        result.append(subresult)
    return result



class UnknownShapeError(RuntimeError):

    def __init__(self, *args, **kwargs):
        RuntimeError.__init__(*args, ** kwargs)


class DataSource(object):

    def get(self, fieldname, frames):
        """
        Returns a NumPy array (or list of arrays) holding the data for a given fieldname.
        The batch size of the array (or list length) is equal to len(frames).

        :param fieldname: stream identifier string
        :param frames: list or tuple of frames
        """
        raise NotImplementedError(self)

    def list_fieldnames(self):
        """
        Returns a list of all fieldnames contained in this source.
        """
        raise NotImplementedError(self)

    def size(self, lookup=False):
        """
        Returns the number of data points in this source.

        If the source is infinite, returns numpy.inf.
        If the size is unknown at this point and lookup=False, returns None.

        :param lookup: If the size should be determined if unknown
        """
        raise NotImplementedError(self)

    def frames(self):
        """
        Returns an iterable object to read all frames of this DataSource.
        The iterator need not provide a length.
        """
        raise NotImplementedError(self)

    def shape(self, fieldname):
        """
        Returns a 1D tensor holding the tensor shape of a single data point from the fieldname.
        The first entry, specifying the batch dimension, must be equal to 1.

        If the shape cannot be determined, this method returns None.

        :param fieldname: stream identifier string
        """
        raise NotImplementedError(self)


class SceneSource(DataSource):

    def __init__(self, scene, frames=None, shape_map=None):
        self.scene = scene
        self._frames = frames
        self._shape_map = shape_map if shape_map is not None else dict()

    def get(self, fieldname, frames):
        for frame in frames:
            yield self.scene.read_array(fieldname, frame)

    def list_fieldnames(self):
        return self.scene.fieldnames

    def size(self, lookup=False):
        if self._frames is None and lookup:
            self._frames = self.scene.frames
        return len(self._frames) if self._frames is not None else None

    def frames(self):
        if self._frames is None:
            self._frames = self.scene.frames
        return self._frames

    def shape(self, fieldname):
        if fieldname in self._shape_map:
            return self._shape_map[fieldname]
        first_frame = next(self.frames())
        first_array = self.get(fieldname, [first_frame])
        shape = first_array.shape
        self._shape_map[fieldname] = shape
        return shape

    def __repr__(self):
        return "SceneSource[%s, frames=%s]" % (self.scene, self._frames)



from phi.geom import AABox
from phi.physics.field import CenteredGrid
from . import tf


def conv_layer(grid, filters, kernel_size, strides=1, padding='valid', activation=None, use_bias=True, name=None, trainable=True, reuse=None):
    assert isinstance(grid, CenteredGrid)
    if grid.rank == 2:
        result = tf.layers.conv2d(grid.data, filters, kernel_size, strides=strides, activation=activation, padding=padding, name=name, use_bias=use_bias, trainable=trainable, reuse=reuse)
    elif grid.rank == 1:
        result = tf.layers.conv1d(grid.data, filters, kernel_size, strides=strides, activation=activation, padding=padding, name=name, use_bias=use_bias, trainable=trainable, reuse=reuse)
    else:
        raise NotImplementedError()
    if padding == 'valid':
        w_upper = kernel_size // 2
        w_lower = (kernel_size - 1) // 2
        box = AABox(grid.box.lower + w_lower * grid.dx, grid.box.upper - w_upper * grid.dx)
    else:
        box = grid.box
    return CenteredGrid(result, box=box, extrapolation=grid.extrapolation)


def residual_block(grid, nb_channels, kernel_size=3, stride=1, activation=tf.nn.leaky_relu, _project_shortcut=False, padding="SYMMETRIC", name=None, training=False, trainable=True, reuse=tf.AUTO_REUSE):
    if grid.rank == 2:
        return _residual_block_2d(grid, nb_channels, kernel_size, stride, activation, _project_shortcut, padding, name, training, trainable, reuse)
    if grid.rank == 1:
        return _residual_block_1d(grid, nb_channels, kernel_size, stride, activation, _project_shortcut, padding, name, training, trainable, reuse)
    raise NotImplementedError()


def _residual_block_1d(grid, nb_channels, kernel_size=3, stride=1, activation=tf.nn.leaky_relu, _project_shortcut=False, padding="SYMMETRIC", name=None, training=False, trainable=True, reuse=tf.AUTO_REUSE):
    y = grid.data
    shortcut = y

    pad1 = [(kernel_size - 1) // 2, kernel_size // 2]

    # down-sampling is performed with a stride of 2
    y = tf.pad(y, [[0, 0], pad1, [0, 0]], mode=padding)
    y = tf.layers.conv1d(y, nb_channels, kernel_size=kernel_size, strides=stride, padding='valid', name=None if name is None else name + "/conv1", trainable=trainable, reuse=reuse)
    # y = tf.layers.batch_normalization(y, name=None if name is None else name+"/norm1", training=training, trainable=trainable, reuse=reuse)
    y = activation(y)

    y = tf.pad(y, [[0, 0], pad1, [0, 0]], mode=padding)
    y = tf.layers.conv1d(y, nb_channels, kernel_size=kernel_size, strides=1, padding='valid', name=None if name is None else name + "/conv2", trainable=trainable, reuse=reuse)
    # y = tf.layers.batch_normalization(y, name=None if name is None else name+"/norm2", training=training, trainable=trainable, reuse=reuse)

    # identity shortcuts used directly when the input and output are of the same dimensions
    if _project_shortcut or stride != 1:
        # when the dimensions increase projection shortcut is used to match dimensions (done by 1×1 convolutions)
        # when the shortcuts go across feature maps of two sizes, they are performed with a stride of 2
        shortcut = tf.pad(shortcut, [[0, 0], pad1, [0, 0]], mode=padding)
        shortcut = tf.layers.conv1d(shortcut, nb_channels, kernel_size=(1, 1), strides=stride, padding='valid', name=None if name is None else name + "/convid", trainable=trainable, reuse=reuse)
        # shortcut = tf.layers.batch_normalization(shortcut, name=None if name is None else name+"/normid", training=training, trainable=trainable, reuse=reuse)

    y += shortcut
    y = activation(y)
    return CenteredGrid(y, box=grid.box, extrapolation=grid.extrapolation)


def _residual_block_2d(grid, nb_channels, kernel_size=(3, 3), stride=(1, 1), activation=tf.nn.leaky_relu, _project_shortcut=False, padding="SYMMETRIC", name=None, training=False, trainable=True, reuse=tf.AUTO_REUSE):
    y = grid.data
    shortcut = y

    if isinstance(kernel_size, int):
        kernel_size = (kernel_size, kernel_size)

    pad1 = [(kernel_size[0] - 1) // 2, kernel_size[0] // 2]
    pad2 = [(kernel_size[1] - 1) // 2, kernel_size[1] // 2]

    # down-sampling is performed with a stride of 2
    y = tf.pad(y, [[0, 0], pad1, pad2, [0, 0]], mode=padding)
    y = tf.layers.conv2d(y, nb_channels, kernel_size=kernel_size, strides=stride, padding='valid', name=None if name is None else name + "/conv1", trainable=trainable, reuse=reuse)
    # y = tf.layers.batch_normalization(y, name=None if name is None else name+"/norm1", training=training, trainable=trainable, reuse=reuse)
    y = activation(y)

    y = tf.pad(y, [[0, 0], pad1, pad2, [0, 0]], mode=padding)
    y = tf.layers.conv2d(y, nb_channels, kernel_size=kernel_size, strides=(1, 1), padding='valid', name=None if name is None else name + "/conv2", trainable=trainable, reuse=reuse)
    # y = tf.layers.batch_normalization(y, name=None if name is None else name+"/norm2", training=training, trainable=trainable, reuse=reuse)

    # identity shortcuts used directly when the input and output are of the same dimensions
    if _project_shortcut or stride != (1, 1):
        # when the dimensions increase projection shortcut is used to match dimensions (done by 1×1 convolutions)
        # when the shortcuts go across feature maps of two sizes, they are performed with a stride of 2
        shortcut = tf.pad(shortcut, [[0, 0], pad1, pad2, [0, 0]], mode=padding)
        shortcut = tf.layers.conv2d(shortcut, nb_channels, kernel_size=(1, 1), strides=stride, padding='valid', name=None if name is None else name + "/convid", trainable=trainable, reuse=reuse)
        # shortcut = tf.layers.batch_normalization(shortcut, name=None if name is None else name+"/normid", training=training, trainable=trainable, reuse=reuse)

    y += shortcut
    y = activation(y)
    return CenteredGrid(y, box=grid.box, extrapolation=grid.extrapolation)



from phi.backend.backend import Backend
from phi.backend.dynamic_backend import DYNAMIC_BACKEND
from phi.backend.scipy_backend import SciPyBackend
from phi.struct.struct_backend import StructBroadcastBackend
from .math_util import types, is_static_shape, zeros, ones, randn, randfreq, interpolate
from .helper import is_scalar, axes, rank
from .nd import (spatial_rank, spatial_dimensions, all_dimensions,
                 indices_tensor,
                 normalize_to,
                 batch_align, batch_align_scalar,
                 blur,
                 l1_loss, l2_loss, l_n_loss, frequency_loss,
                 divergence, gradient, axis_gradient, laplace,
                 fourier_laplace, fourier_poisson, fftfreq, abs_square,
                 downsample2x, upsample2x, interpolate_linear,
                 spatial_sum,)
from .batched import BATCHED, ShapeMismatch
from . import optim


# Setup Backend
DYNAMIC_BACKEND.add_backend(SciPyBackend())
DYNAMIC_BACKEND.add_backend(StructBroadcastBackend(DYNAMIC_BACKEND))


def set_precision(floating_point_bits):
    """
    Sets the floating point precision of DYNAMIC_BACKEND which affects all registered backends.

    If `floating_point_bits` is an integer, all floating point tensors created henceforth will be of the corresponding data type, float16, float32 or float64.
    Operations may also convert floating point values to this precision, even if the input had a different precision.

    If `floating_point_bits` is None, new tensors will default to float32 unless specified otherwise.
    The output of math operations has the same precision as its inputs.

    :param floating_point_bits: one of (16, 32, 64, None)
    """
    DYNAMIC_BACKEND.precision = floating_point_bits


# Enable importing methods directly from math
choose_backend = DYNAMIC_BACKEND.choose_backend

abs = DYNAMIC_BACKEND.abs
add = DYNAMIC_BACKEND.add
all = DYNAMIC_BACKEND.all
any = DYNAMIC_BACKEND.any
as_tensor = DYNAMIC_BACKEND.as_tensor
batch_gather = DYNAMIC_BACKEND.batch_gather
boolean_mask = DYNAMIC_BACKEND.boolean_mask
cast = DYNAMIC_BACKEND.cast
ceil = DYNAMIC_BACKEND.ceil
clip = DYNAMIC_BACKEND.clip
copy = DYNAMIC_BACKEND.copy
cos = DYNAMIC_BACKEND.cos
concat = DYNAMIC_BACKEND.concat
conv = DYNAMIC_BACKEND.conv
dimrange = DYNAMIC_BACKEND.dimrange
div = DYNAMIC_BACKEND.div
divide_no_nan = DYNAMIC_BACKEND.divide_no_nan
dot = DYNAMIC_BACKEND.dot
dtype = DYNAMIC_BACKEND.dtype
einsum = DYNAMIC_BACKEND.einsum
equal = DYNAMIC_BACKEND.equal
exp = DYNAMIC_BACKEND.exp
expand_dims = DYNAMIC_BACKEND.expand_dims
fft = DYNAMIC_BACKEND.fft
flatten = DYNAMIC_BACKEND.flatten
floor = DYNAMIC_BACKEND.floor
gather = DYNAMIC_BACKEND.gather
gather_nd = DYNAMIC_BACKEND.gather_nd
ifft = DYNAMIC_BACKEND.ifft
imag = DYNAMIC_BACKEND.imag
isfinite = DYNAMIC_BACKEND.isfinite
is_tensor = DYNAMIC_BACKEND.is_tensor
matmul = DYNAMIC_BACKEND.matmul
max = DYNAMIC_BACKEND.max
maximum = DYNAMIC_BACKEND.maximum
mean = DYNAMIC_BACKEND.mean
min = DYNAMIC_BACKEND.min
minimum = DYNAMIC_BACKEND.minimum
mul = DYNAMIC_BACKEND.mul
name = DYNAMIC_BACKEND.name
ndims = DYNAMIC_BACKEND.ndims
ones_like = DYNAMIC_BACKEND.ones_like
pad = DYNAMIC_BACKEND.pad
pow = DYNAMIC_BACKEND.pow
py_func = DYNAMIC_BACKEND.py_func
random_uniform = DYNAMIC_BACKEND.random_uniform
range = DYNAMIC_BACKEND.range
real = DYNAMIC_BACKEND.real
resample = DYNAMIC_BACKEND.resample
reshape = DYNAMIC_BACKEND.reshape
round = DYNAMIC_BACKEND.round
sign = DYNAMIC_BACKEND.sign
size = DYNAMIC_BACKEND.size
scatter = DYNAMIC_BACKEND.scatter
shape = DYNAMIC_BACKEND.shape
sin = DYNAMIC_BACKEND.sin
sparse_tensor = DYNAMIC_BACKEND.sparse_tensor
sqrt = DYNAMIC_BACKEND.sqrt
stack = DYNAMIC_BACKEND.stack
staticshape = DYNAMIC_BACKEND.staticshape
std = DYNAMIC_BACKEND.std
sub = DYNAMIC_BACKEND.sub
sum = DYNAMIC_BACKEND.sum
prod = DYNAMIC_BACKEND.prod
tile = DYNAMIC_BACKEND.tile
to_complex = DYNAMIC_BACKEND.to_complex
to_float = DYNAMIC_BACKEND.to_float
to_int = DYNAMIC_BACKEND.to_int
unstack = DYNAMIC_BACKEND.unstack
where = DYNAMIC_BACKEND.where
while_loop = DYNAMIC_BACKEND.while_loop
with_custom_gradient = DYNAMIC_BACKEND.with_custom_gradient
zeros_like = DYNAMIC_BACKEND.zeros_like



import os
import numpy as np
import tensorflow as tf
from tensorflow.python.framework import ops

from phi.struct.tensorop import collapsed_gather_nd

# Load Custom Ops
librariesLoaded = False
try:
    current_dir = os.path.dirname(os.path.realpath(__file__))
    resample_op_path = os.path.join(current_dir, 'cuda/build/resample.so')
    resample_gradient_op_path = os.path.join(current_dir, 'cuda/build/resample_gradient.so')
    assert os.path.isfile(
        resample_op_path), 'CUDA binaries not found at %s. Run "python setup.py cuda" to compile them' % resample_op_path
    assert os.path.isfile(resample_gradient_op_path), 'CUDA binaries not found at %s. Run "python setup.py cuda" to ' \
                                                      'compile them' % resample_gradient_op_path
    resample_op = tf.load_op_library(resample_op_path)
    resample_gradient_op = tf.load_op_library(resample_gradient_op_path)
    librariesLoaded = True
except (RuntimeError, AssertionError) as e:
    print('Could not load resample cuda libraries:', e)
    librariesLoaded = False

# Register gradient


@ops.RegisterGradient("Resample")
def _resample_gradient(op, gradient):
    gradients = resample_gradient_op.resample_gradient(gradient, op.inputs[0], op.inputs[1], op.inputs[2])
    return [gradients[0], gradients[1], None]


def use_cuda(inputs):
    if not librariesLoaded:
        return False
    if not tf.test.is_gpu_available(True, (3, 0)):
        return False
    shape = inputs.shape
    dims = len(shape) - 2
    components = shape[len(shape) - 1]
    if dims > 3 or components > 4:
        return False
    if dims == 1 and shape[1] > 8192:
        return False
    if dims == 2 and (shape[1] > 32768 or shape[2] > 65536):
        return False
    if dims == 3 and (shape[1] > 2048 or shape[2] > 2048 or shape[3] > 2048):
        return False
    return True


def resample_cuda(inputs, sample_coords, boundary):
    ZERO = 0
    REPLICATE = 1
    CIRCULAR = 2
    SYMMETRIC = 3
    REFLECT = 4
    shape = inputs.shape
    dims = len(shape) - 2
    boundary_array = np.zeros((dims, 2), np.uint32)
    for i in range(dims):
        for j in range(2):
            current_boundary = collapsed_gather_nd(boundary, [i, j]).lower()
            if current_boundary == 'zero' or current_boundary == 'constant':
                boundary_array[i, j] = ZERO
            elif current_boundary == 'replicate':
                boundary_array[i, j] = REPLICATE
            elif current_boundary == 'circular' or current_boundary == 'wrap':
                boundary_array[i, j] = CIRCULAR
            elif current_boundary == 'symmetric':
                boundary_array[i, j] = SYMMETRIC
            elif current_boundary == 'reflect':
                boundary_array[i, j] = REFLECT

    return resample_op.resample(inputs, sample_coords, boundary_array)



from phi import math
from .solver_api import PoissonSolver


class FourierSolver(PoissonSolver):

    def __init__(self):
        """
        Computes the inverse laplace operation in Fourier space.

        This is computationally inexpensive compared to iterative solvers; the FFT is the most expensive step.

        While the result is only correct for periodic domains, it can be used as initial guess for other solvers, even for non-periodic domains.
        """
        PoissonSolver.__init__(self, 'FFT', ('CPU', 'GPU'), supports_guess=False, supports_loop_counter=False, supports_continuous_masks=False)

    def solve(self, field, domain, guess, enable_backprop):
        return math.fourier_poisson(field), None



import six


class Trait(object):
    """
Traits are always handled in the order in which they were declared.
Inherited traits come before newly declared structs, unless explicitly re-declared.
    """

    def __init__(self, keywords=()):
        self.keywords = keywords
        for keyword in keywords:
            assert isinstance(keyword, six.string_types)

    def check_argument(self, struct_class, item, keyword, value):
        """
Called when a keyword of the trait is used on an item.
        :param struct_class: struct class
        :param item: Item (usually created via @variable or @constant)
        :param keyword: keyword present on item and part of the trait (string)
        :param value: value associated with the keyword for the given item
        """
        pass

    def endow(self, struct):
        """
Called on newly created Structs with this trait.
This method is called before the first validation.
        :param struct: struct instance to be endowed with the trait
        """
        pass

    def pre_validate_struct(self, struct):
        """
Called before a struct instance with the trait is validated.
        :param struct: struct about to be validated
        """
        pass

    def post_validate_struct(self, struct):
        """
Called after a struct instance with the trait is validated.
        :param struct: validated struct
        """
        pass

    def pre_validated(self, struct, item, value):
        """
Processes the value of an item before the validation function is called.
        :param struct: struct undergoing validation
        :param item: item being validated
        :param value: item value before validation
        :return: processed item value which will be passed to the validation function
        """
        return value

    def post_validated(self, struct, item, value):
        """
Processes the value of an item after the validation function is called.
        :param struct: struct undergoing validation
        :param item: item being validated
        :param value: item value after validation
        :return: processed item value which will be stored in the struct
        """
        return value

    def __eq__(self, other):
        return self.__class__ == other.__class__ if isinstance(other, Trait) else False

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return '%s (Trait)' % self.__class__.__name__

    def __hash__(self):
        return hash(self.__class__)



# coding=utf-8
import warnings

from ..backend.dynamic_backend import DYNAMIC_BACKEND as math
from .optim import conjugate_gradient as new_cg


def conjugate_gradient(k, apply_A, initial_x=None, accuracy=1e-5, max_iterations=1024, back_prop=False):
    warnings.warn("conjugate_gradient from phi.math.blas is deprecated. Use phi.math.optim.conjugate_gradient instead.", DeprecationWarning)
    if initial_x is None:
        initial_x = math.zeros_like(k)
    result = new_cg(function=apply_A, y=k, x0=initial_x, accuracy=accuracy, max_iterations=max_iterations, back_prop=back_prop)
    return result.x, result.iterations



import warnings

import numpy as np
from phi import math, struct
from phi.geom import AABox
from phi.geom.geometry import assert_same_rank
from phi.struct.tensorop import collapse, collapsed_gather_nd

from . import State
from .material import OPEN, Material


@struct.definition()
class Domain(struct.Struct):

    def __init__(self, resolution, boundaries=OPEN, box=None, **kwargs):
        """
        Simulation domain that specifies size and boundary conditions.

        If all boundary surfaces should have the same behaviour, pass a single Material instance.

        To specify the boundary constants_dict per dimension or surface, pass a tuple or list with as many elements as there are spatial dimensions (highest dimension first).
        Each element can either be a Material, specifying the faces perpendicular to that axis, or a pair
        of Material holding (lower_face_material, upper_face_material).

        Examples:

        Domain(grid, OPEN) - all surfaces are open

        DomainBoundary(grid, boundaries=[(SLIPPY, OPEN), SLIPPY]) - creates a 2D domain with an open top and otherwise solid boundaries

        :param grid: Grid object or 1D tensor specifying the grid dimensions
        :param boundaries: Material or list of Material/Pair of Material
        """
        struct.Struct.__init__(self, **struct.kwargs(locals()))

    @staticmethod
    def as_domain(domain_like):
        assert domain_like is not None
        if isinstance(domain_like, Domain):
            return domain_like
        if isinstance(domain_like, int):
            return Domain([domain_like])
        if isinstance(domain_like, (tuple, list)):
            return Domain(domain_like)
        raise ValueError('Not a valid domain: %s' % domain_like)

    @struct.constant()
    def resolution(self, resolution):
        if len(math.staticshape(resolution)) == 0:
            resolution = [resolution]
        return np.array(resolution)

    @struct.constant(dependencies='resolution')
    def box(self, box):
        return AABox.to_box(box, resolution_hint=self.resolution)

    @struct.derived()
    def dx(self):
        return self.box.size / self.resolution

    @struct.constant(default=OPEN)
    def boundaries(self, boundaries):
        assert isinstance(boundaries, (Material, list, tuple))
        if isinstance(boundaries, (tuple, list)):
            assert len(boundaries) == self.rank
        return collapse(boundaries)

    @property
    def rank(self):
        return len(self.resolution)

    def __repr__(self):
        if self.is_valid:
            return '(%s, size=%s)' % (self.resolution, self.box.size)
        else:
            return struct.Struct.__repr__(self)

    def cell_index(self, global_position):
        local_position = self.box.global_to_local(global_position) * self.resolution
        position = math.to_int(local_position - 0.5)
        position = math.maximum(0, position)
        position = math.minimum(position, self.resolution - 1)
        return position

    def center_points(self):
        from .field import CenteredGrid
        warnings.warn("Domain.center_points is deprecated. Use CenteredGrid.getpoints().data instead.", DeprecationWarning)
        return CenteredGrid.getpoints(self.box, self.resolution).data
        # idx_zyx = np.meshgrid(*[np.arange(0.5, dim + 0.5, 1) for dim in self.resolution], indexing="ij")
        # return math.expand_dims(math.stack(idx_zyx, axis=-1), 0)

    def staggered_points(self, dimension):
        idx_zyx = np.meshgrid(*[np.arange(0.5, dim + 1.5, 1) if dim != dimension else np.arange(0, dim + 1, 1) for dim in self.resolution], indexing="ij")
        return math.expand_dims(math.stack(idx_zyx, axis=-1), 0)

    def indices(self):
        """
        Constructs a grid containing the index-location as components.
        Each index denotes the location within the tensor starting from zero.
        Indices are encoded as vectors in the index tensor.

        :param dtype: a numpy data type (default float32)
        :return: an index tensor of shape (1, spatial dimensions..., spatial rank)
        """
        idx_zyx = np.meshgrid(*[range(dim) for dim in self.resolution], indexing="ij")
        return math.expand_dims(np.stack(idx_zyx, axis=-1))

    @staticmethod
    def equal(grid1, grid2):
        assert isinstance(grid1, Domain), 'Not a Domain: %s' % type(grid1)
        assert isinstance(grid2, Domain), 'Not a Domain: %s' % type(grid2)
        return np.all(grid1.resolution == grid2.resolution) and grid1.box == grid2.box

    def centered_shape(self, components=1, batch_size=1, name=None, extrapolation=None, age=0.0):
        warnings.warn("Domain.centered_shape and Domain.centered_grid are deprecated. Use CenteredGrid.sample() instead.", DeprecationWarning)
        from phi.physics.field import CenteredGrid
        return CenteredGrid(tensor_shape(batch_size, self.resolution, components), age=age, box=self.box, extrapolation=extrapolation, name=name, batch_size=batch_size, flags=(), content_type=struct.Struct.shape)

    def staggered_shape(self, batch_size=1, name=None, extrapolation=None, age=0.0):
        grids = []
        for axis in range(self.rank):
            shape = _extend1(tensor_shape(batch_size, self.resolution, 1), axis)
            from phi.physics.field.staggered_grid import staggered_component_box
            box = staggered_component_box(self.resolution, axis, self.box)
            from phi.physics.field import CenteredGrid
            grid = CenteredGrid(shape, box, age=age, extrapolation=extrapolation, name=None, batch_size=batch_size, flags=(), content_type=struct.Struct.shape)
            grids.append(grid)
        from phi.physics.field import StaggeredGrid
        return StaggeredGrid(grids, age=age, box=self.box, name=name, batch_size=batch_size, extrapolation=extrapolation, flags=(), content_type=struct.Struct.shape)

    def centered_grid(self, data, components=1, dtype=None, name=None, batch_size=None, extrapolation=None):
        warnings.warn("Domain.centered_shape and Domain.centered_grid are deprecated. Use CenteredGrid.sample() instead.", DeprecationWarning)
        from phi.physics.field import CenteredGrid
        if callable(data):  # data is an initializer
            shape = self.centered_shape(components, batch_size=batch_size, name=name, extrapolation=extrapolation, age=())
            try:
                grid = data(shape, dtype=dtype)
            except TypeError:
                grid = data(shape)
            if grid.age == ():
                grid._age = 0.0
        else:
            grid = CenteredGrid.sample(data, self, batch_size=batch_size)
        assert grid.component_count == components, "Field has %d components but %d are required for '%s'" % (grid.component_count, components, name)
        if dtype is not None and math.dtype(grid.data) != dtype:
            grid = grid.copied_with(data=math.cast(grid.data, dtype))
        if name is not None:
            grid = grid.copied_with(name=name, tags=(name,) + grid.tags)
        if extrapolation is not None:
            grid = grid.copied_with(extrapolation=extrapolation)
        return grid

    def _centered_grid(self, data, components=1, dtype=None, name=None, batch_size=None, extrapolation=None):
        warnings.warn("Domain.centered_shape and Domain.centered_grid are deprecated. Use CenteredGrid.sample() instead.", DeprecationWarning)
        from phi.physics.field import CenteredGrid
        if extrapolation is None:
            extrapolation = Material.extrapolation_mode(self.boundaries)
        if callable(data):  # data is an initializer
            shape = self.centered_shape(components, batch_size=batch_size, name=name, extrapolation=extrapolation, age=())
            try:
                data = data(shape, dtype=dtype)
            except TypeError:
                data = data(shape)
            if data.age == ():
                data._age = 0.0
        from phi.physics.field import Field
        if isinstance(data, Field):
            assert_same_rank(data.rank, self.rank, 'data does not match Domain')
            data = data.at(CenteredGrid.getpoints(self.box, self.resolution))
            if name is not None:
                data = data.copied_with(name=name, extrapolation=extrapolation)
                data._batch_size = batch_size
            grid = data
        elif isinstance(data, (int, float)):
            shape = self.centered_shape(components, batch_size=batch_size, name=name, extrapolation=extrapolation, age=0.0)
            grid = math.zeros(shape, dtype=dtype) + data
        else:
            grid = CenteredGrid(data, box=self.box, extrapolation=extrapolation, name=name)
        return grid

    def staggered_grid(self, data, dtype=None, name=None, batch_size=None, extrapolation=None):
        if extrapolation is None:
            extrapolation = Material.extrapolation_mode(self.boundaries)
        if callable(data):  # data is an initializer
            shape = self.staggered_shape(batch_size=batch_size, name=name, extrapolation=extrapolation, age=())
            try:
                data = data(shape, dtype=dtype)
            except TypeError:
                data = data(shape)
            if data.age == ():
                data._age = 0.0
                for field in data.data:
                    field._age = 0.0
        from phi.physics.field import Field
        if isinstance(data, Field):
            from phi.physics.field import StaggeredGrid
            if isinstance(data, StaggeredGrid) and np.all(data.resolution == self.resolution) and data.box == self.box:
                grid = data
            else:
                grid = data.at(StaggeredGrid.sample(0, self, batch_size=batch_size))  # ToDo this is not ideal
        elif isinstance(data, (int, float)):
            shape = self.staggered_shape(batch_size=batch_size, name=name, extrapolation=extrapolation)
            from phi.physics.field import DIVERGENCE_FREE
            grid = (math.zeros(shape, dtype=dtype) + data).copied_with(flags=[DIVERGENCE_FREE])
        else:
            from .field import StaggeredGrid
            grid = StaggeredGrid(data, self.box, name, batch_size=None, extrapolation=extrapolation)
        return grid

    def surface_material(self, axis=0, upper_boundary=False):
        return collapsed_gather_nd(self.boundaries, axis, upper_boundary)


def _friction_mask(masks_and_multipliers):
    for mask, multiplier in masks_and_multipliers:
        return mask


def tensor_shape(batch_size, resolution, components):
    return np.concatenate([[batch_size], resolution, [components]])


def _extend1(shape, axis):
    shape = list(shape)
    shape[axis + 1] += 1
    return shape


@struct.definition()
class DomainState(State):

    @struct.constant()
    def domain(self, domain):
        return Domain.as_domain(domain)

    @property
    def resolution(self):
        return self.domain.resolution

    @property
    def rank(self):
        return self.domain.rank

    def centered_grid(self, name, value, components=1, dtype=None):
        extrapolation = Material.extrapolation_mode(self.domain.boundaries)
        return self.domain.centered_grid(value, dtype=dtype, name=name, components=components, batch_size=self._batch_size, extrapolation=extrapolation)

    def staggered_grid(self, name, value, dtype=None):
        extrapolation = Material.vector_extrapolation_mode(self.domain.boundaries)
        return self.domain.staggered_grid(value, dtype=dtype, name=name, batch_size=self._batch_size, extrapolation=extrapolation)



# coding=utf-8
import numpy

import dash
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output
from dash.exceptions import PreventUpdate

from phi.app import App
from phi import struct


class DashApp:

    def __init__(self, app, config, header_layout):
        assert isinstance(app, App)
        self.app = app
        self.config = config
        self.dash = dash.Dash(u'PhiFlow')
        self.dash.config.suppress_callback_exceptions = True
        self.hrefs = set()
        self.page_urls = {}
        self.field_minmax = {}
        self.minmax_decay = 0.975

        # The index page encapsulates the specific pages.
        self.dash.layout = html.Div([
            dcc.Location(id='url', refresh=False),
            header_layout,
            html.Div(id='page-content')  # Content is set using the URL
        ], style={'fontFamily': 'arial'})

        @self.dash.callback(Output('page-content', 'children'), [Input('url', 'pathname')])
        def display_page(pathname):
            if pathname in self.page_urls:
                layout = self.page_urls[pathname]
                if callable(layout):
                    return layout(self)
                else:
                    return layout
            else:
                return html.Div([
                    html.Div('404 - No such page: %s' % pathname),
                    dcc.Link('Back to main page', href='/'),
                ])

    @property
    def server(self):
        return self.dash.server

    def search_callback(self, output, hrefs):
        assert isinstance(output, Output)
        self.hrefs.update(hrefs)

        def decorator(func):
            @self.dash.callback(output, [Input('url', 'search')])
            def href_callback(search):
                if search in hrefs:
                    func(search)
                else:
                    raise PreventUpdate()
        return decorator

    def consumes(self, href):
        return href in self.hrefs

    def add_page(self, path, page_layout):
        self.page_urls[path] = page_layout

    def get_field(self, name):
        data = self.app.get_field(name)
        if data is not None:
            tensors = struct.flatten(data)
            if tensors:
                minimum = min([numpy.min(tensor) for tensor in tensors])
                maximum = max([numpy.max(tensor) for tensor in tensors])
                if name in self.field_minmax:
                    midpoint = (self.field_minmax[name][1] + self.field_minmax[name][0]) / 2
                    minimum = min(minimum, self.field_minmax[name][0] * self.minmax_decay + midpoint * (1 - self.minmax_decay))
                    maximum = max(maximum, self.field_minmax[name][1] * self.minmax_decay + midpoint * (1 - self.minmax_decay))
                self.field_minmax[name] = (minimum, maximum)
        return data

    def reset_field_summary(self):
        self.field_minmax = {}

    def get_minmax(self, field):
        if field in self.field_minmax:
            return self.field_minmax[field]
        else:
            return 0, 0



import six

from .context import _struct_context, _STRUCT_CONTEXT_STACK
from .structdef import Item


class ItemCondition(object):
    """
ItemConditions are used to filter struct items.
They represent a named boolean function on items.

In addition, they can be used in 'with ItemCondition:' blocks, adding the condition to the thread context for all actions within that block.
In particular, struct.map, Struct.shape, Struct.staticshape are affected by context conditions.

This module provides some standard conditions like ALL_ITEMS, DATA, VARIABLES, CONSTANTS.
    """

    def __init__(self, item_condition, name=None):
        assert item_condition is None or callable(item_condition), item_condition
        self.item_condition = item_condition
        if name is not None:
            self.name = name
        else:
            self.name = item_condition.__name__ if item_condition is not None else 'ALL'

    def condition_check(self, item):
        return True if self.item_condition is None else self.item_condition(item)

    __call__ = condition_check

    def __enter__(self):
        self.context = _struct_context(self)
        return self.context.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        result = self.context.__exit__(exc_type, exc_val, exc_tb)
        self.context = None
        return result

    def __repr__(self):
        return self.name

    def __and__(self, other):
        assert isinstance(other, ItemCondition)
        return ItemCondition(lambda item: self(item) and other(item), name='%s and %s' % (self.name, other.name))

    def __or__(self, other):
        assert isinstance(other, ItemCondition)
        return ItemCondition(lambda item: self(item) or other(item), name='%s or %s' % (self.name, other.name))


CONSTANTS = ItemCondition(lambda item: not item.is_variable, 'CONSTANTS')
VARIABLES = ItemCondition(lambda item: item.is_variable, 'VARIABLES')
DATA = ItemCondition(lambda item: item.holds_data, 'DATA')
ALL_ITEMS = ItemCondition(None)


def context_item_condition(item):
    """
Checks all thread-global item conditions.
Conditions can be specified using 'with ItemCondition:' blocks.
If no condition was specified, this function defaults to testing whether the item holds data.
    :param item: item to be checked
    :return: True if the item passes all conditions, False otherwise
    """
    user_specified = False
    for context in _STRUCT_CONTEXT_STACK:
        if isinstance(context, ItemCondition):
            user_specified = True
            if not context.condition_check(item):
                return False
    if user_specified:
        return True
    else:
        return item.holds_data  # default condition if none specified


def ignore(items):
    if not isinstance(items, (tuple, list)):
        items = (items,)
    for ignored_item in items:
        assert isinstance(ignored_item, Item) or isinstance(ignored_item, six.string_types)

    def is_ignored(item):
        for ignored in items:
            if item is ignored:
                return True
            if item.name == ignored:
                return True
        return False

    condition = ItemCondition(lambda item: not is_ignored(item), 'ignore %s' % items)
    return condition
